import os
import ast
import configparser


config = {}

config_file = os.path.join(os.path.dirname(__file__), 'config.ini')

config_parser = configparser.ConfigParser()

config_parser.read(config_file)

ASSOCIATIONS = {}

for section in config_parser.sections():
    for k, v in config_parser[section].items():
        ASSOCIATIONS[k] = section


def _update_config(config_parser, config_dict):

    for section in config_parser.sections():

        for k, v in config_parser[section].items():

            if v in ['True', 'False']:
                config_dict[k] = True if v == 'True' else False

            elif v == 'None':
                config_dict[k] = None
            else:

                try:
                    config_dict[k] = ast.literal_eval(v)
                except:
                    config_dict[k] = v

    return config_dict


config = _update_config(config_parser, config)


class update(object):

    """
    >>> with gw.config.update(sensor='l8'):
    >>>
    >>>     with gw.open('image.tif') as ds:
    >>>         print(ds.gw.config)
    """

    def __init__(self, config=config, **kwargs):

        self.config = config
        self._set_defaults(config)

        if kwargs:
            self._assign(config, **kwargs)

    def __enter__(self):
        return self.config

    def __exit__(self, type, value, traceback):
        d = self.config
        self._set_defaults(d)

    @staticmethod
    def _set_defaults(d):

        config_parser.read(config_file)
        d = _update_config(config_parser, d)

    @staticmethod
    def _assign(d, **kwargs):

        if kwargs:

            for k, v in kwargs.items():
                config_parser[ASSOCIATIONS[k]][k] = str(v)

        d = _update_config(config_parser, d)



import os

rgbn = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'rgbn.tif')
rgbn_suba = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'rgbn_suba.tif')
rgbn_subb = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'rgbn_subb.tif')

rgbn_20160101 = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'rgbn_20160101.tif')
rgbn_20160401 = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'rgbn_20160401.tif')
rgbn_20160517 = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'rgbn_20160517.tif')
rgbn_20170203 = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'rgbn_20170203.tif')

rgbn_time_list = [rgbn_20160101, rgbn_20160401, rgbn_20160517, rgbn_20170203]



import os
import shutil
from collections import namedtuple

from ..errors import logger

import numpy as np
import rasterio as rio
from rasterio.crs import CRS
from rasterio.enums import Resampling
from rasterio.vrt import WarpedVRT
from rasterio.warp import aligned_target, calculate_default_transform, transform_bounds, reproject
from rasterio.transform import array_bounds, from_bounds
from rasterio.windows import Window
from rasterio.coords import BoundingBox

from affine import Affine
import zarr
import numcodecs


class WriteDaskArray(object):

    """
    ``Rasterio`` wrapper to allow ``dask.array.store`` to save chunks as windows.

    Args:
        filename (str): The file to write to.
        overwrite (Optional[bool]): Whether to overwrite an existing output file.
        separate (Optional[bool]): Whether to write blocks as separate files. Otherwise, write to the same file.
        out_block_type (Optional[str]): The output block type. Choices are ['GTiff', 'zarr'].
            *Only used if ``separate`` = ``True``.
        keep_blocks (Optional[bool]): Whether to keep the blocks stored on disk.
            *Only used if ``separate`` = ``True``.
        gdal_cache (Optional[int]): The GDAL cache size (in MB).
        kwargs (Optional[dict]): Other keyword arguments passed to ``rasterio``.

    Reference:
        Code modified from https://github.com/dymaxionlabs/dask-rasterio

    Returns:
        None
    """

    def __init__(self,
                 filename,
                 overwrite=False,
                 separate=False,
                 out_block_type='zarr',
                 keep_blocks=False,
                 gdal_cache=512,
                 **kwargs):

        self.filename = filename
        self.overwrite = overwrite
        self.separate = separate
        self.out_block_type = out_block_type
        self.keep_blocks = keep_blocks
        self.gdal_cache = gdal_cache
        self.kwargs = kwargs

        self.d_name, f_name = os.path.split(self.filename)
        self.f_base, self.f_ext = os.path.splitext(f_name)

        self.root = None
        self.compressor = None
        self.sub_dir = None
        self.zarr_file = None

        if self.separate:

            if self.out_block_type.lower() not in ['gtiff', 'zarr']:

                logger.warning('  The output block type is not recognized. Save blocks as zarr files.')
                self.out_block_type = 'zarr'

            self.sub_dir = os.path.join(self.d_name, 'sub_tmp_')
            self.zarr_file = os.path.join(self.sub_dir, 'data.zarr')

            self.compressor = numcodecs.Blosc(cname='zstd',
                                              clevel=3,
                                              shuffle=numcodecs.Blosc.BITSHUFFLE)

            if os.path.isdir(self.sub_dir):
                shutil.rmtree(self.sub_dir)

            os.makedirs(self.sub_dir)

    def __setitem__(self, key, item):

        if len(key) == 3:

            index_range, y, x = key
            indexes = list(range(index_range.start + 1, index_range.stop + 1, index_range.step or 1))

        else:

            indexes = 1
            y, x = key

        if self.separate:

            if self.out_block_type.lower() == 'zarr':

                group_name = '{BASE}_y{Y:09d}_x{X:09d}_h{H:09d}_w{W:09d}'.format(BASE=self.f_base,
                                                                                 Y=y.start,
                                                                                 X=x.start,
                                                                                 H=y.stop - y.start,
                                                                                 W=x.stop - x.start)

                group = self.root.create_group(group_name)

                z = group.array('data',
                                item,
                                compressor=self.compressor,
                                dtype=item.dtype.name,
                                chunks=(self.kwargs['blockysize'], self.kwargs['blockxsize']))

                group.attrs['row_off'] = y.start
                group.attrs['col_off'] = x.start
                group.attrs['height'] = y.stop - y.start
                group.attrs['width'] = x.stop - x.start

            else:

                out_filename = os.path.join(self.sub_dir,
                                            '{BASE}_y{Y:09d}_x{X:09d}_h{H:09d}_w{W:09d}{EXT}'.format(BASE=self.f_base,
                                                                                                     Y=y.start,
                                                                                                     X=x.start,
                                                                                                     H=y.stop - y.start,
                                                                                                     W=x.stop - x.start,
                                                                                                     EXT=self.f_ext))

                if self.overwrite:

                    if os.path.isfile(out_filename):
                        os.remove(out_filename)

                io_mode = 'w'

                # Every block starts at (0, 0) for the output
                w = Window(col_off=0,
                           row_off=0,
                           width=x.stop - x.start,
                           height=y.stop - y.start)

                # xres, 0, minx, 0, yres, maxy
                # TODO: hardcoded driver type
                kwargs = dict(driver=self.kwargs['driver'],
                              width=w.width,
                              height=w.height,
                              count=self.kwargs['count'],
                              dtype=self.kwargs['dtype'],
                              nodata=self.kwargs['nodata'],
                              blockxsize=self.kwargs['blockxsize'],
                              blockysize=self.kwargs['blockysize'],
                              crs=self.kwargs['crs'],
                              transform=Affine(self.kwargs['transform'][0],
                                               0.0,
                                               self.kwargs['transform'][2] + (x.start * self.kwargs['transform'][0]),
                                               0.0,
                                               self.kwargs['transform'][4],
                                               self.kwargs['transform'][5] - (y.start * -self.kwargs['transform'][4])),
                              compress=self.kwargs['compress'] if 'compress' in self.kwargs else 'none',
                              tiled=True)

        else:

            out_filename = self.filename

            w = Window(col_off=x.start,
                       row_off=y.start,
                       width=x.stop - x.start,
                       height=y.stop - y.start)

            io_mode = 'r+'
            kwargs = {}

        if not self.separate or (self.separate and self.out_block_type.lower() == 'gtiff'):

            with rio.open(out_filename,
                          mode=io_mode,
                          sharing=False,
                          **kwargs) as dst_:

                dst_.write(item,
                           window=w,
                           indexes=indexes)

    def __enter__(self):

        if self.separate:

            if self.out_block_type.lower() == 'zarr':
                self.root = zarr.open(self.zarr_file, mode='w')

        else:

            if 'compress' in self.kwargs:
                logger.warning('\nCannot write concurrently to a compressed raster when using a combination of processes and threads.\nTherefore, compression will be applied after the initial write.')
                del self.kwargs['compress']

            # An alternative here is to leave the writeable object open as self.
            # However, this does not seem to work when used within a Dask
            #   client environment because the `self.dst_` object cannot be pickled.

            # Create the output file
            with rio.open(self.filename, mode='w', **self.kwargs) as dst_:
                pass

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass


def check_res(res):

    """
    Checks a resolution

    Args:
        res (int | float | tuple): The resolution.

    Returns:
        ``tuple``
    """

    if isinstance(res, tuple):
        dst_res = res
    elif isinstance(res, int) or isinstance(res, float):
        dst_res = (res, res)
    else:
        logger.exception('  The resolution should be given as an integer, float, or tuple.')
        raise TypeError

    return dst_res


def check_crs(crs):

    """
    Checks a CRS instance

    Args:
        crs (``CRS`` | int | dict | str): The CRS instance.

    Returns:
        ``rasterio.crs.CRS``
    """

    if isinstance(crs, CRS):
        dst_crs = crs
    elif isinstance(crs, int):
        dst_crs = CRS.from_epsg(crs)
    elif isinstance(crs, dict):
        dst_crs = CRS.from_dict(crs)
    elif isinstance(crs, str):

        if crs.startswith('+proj'):
            dst_crs = CRS.from_proj4(crs)
        else:
            dst_crs = CRS.from_string(crs)

    else:
        logger.exception('  The CRS was not understood.')
        raise TypeError

    return dst_crs


def align_bounds(minx, miny, maxx, maxy, res):

    """
    Aligns bounds to resolution

    Args:
        minx (float)
        miny (float)
        maxx (float)
        maxy (float)
        res (tuple)

    Returns:
        ``affine.Affine``, ``int``, ``int``
    """

    xres, yres = res

    new_height = (maxy - miny) / yres
    new_width = (maxx - minx) / xres

    new_transform = Affine(xres, 0.0, minx, 0.0, -yres, maxy)

    return aligned_target(new_transform, new_width, new_height, res)


def get_file_bounds(filenames,
                    bounds_by='union',
                    crs=None,
                    res=None,
                    return_bounds=False):

    """
    Gets the union of all files

    Args:
        filenames (list): The file names to mosaic.
        bounds_by (Optional[str]): How to concatenate the output extent. Choices are ['intersection', 'union'].
        crs (Optional[crs]): The CRS to warp to.
        res (Optional[tuple]): The cell resolution to warp to.
        return_bounds (Optional[bool]): Whether to return the bounds tuple.

    Returns:
        transform, width, height
    """

    with rio.open(filenames[0]) as src:

        if crs:
            dst_crs = check_crs(crs)
        else:
            dst_crs = src.crs

        if res:
            dst_res = check_res(res)
        else:
            dst_res = src.res

        # Transform the extent to the reference CRS
        bounds_left, bounds_bottom, bounds_right, bounds_top = transform_bounds(src.crs,
                                                                                dst_crs,
                                                                                src.bounds.left,
                                                                                src.bounds.bottom,
                                                                                src.bounds.right,
                                                                                src.bounds.top,
                                                                                densify_pts=21)

    if bounds_by.lower() in ['union', 'intersection']:

        for fn in filenames[1:]:

            with rio.open(fn) as src:

                # Transform the extent to the reference CRS
                left, bottom, right, top = transform_bounds(src.crs,
                                                            dst_crs,
                                                            src.bounds.left,
                                                            src.bounds.bottom,
                                                            src.bounds.right,
                                                            src.bounds.top,
                                                            densify_pts=21)

            # Update the mosaic bounds
            if bounds_by.lower() == 'union':

                bounds_left = min(bounds_left, left)
                bounds_bottom = min(bounds_bottom, bottom)
                bounds_right = max(bounds_right, right)
                bounds_top = max(bounds_top, top)

            elif bounds_by.lower() == 'intersection':

                bounds_left = max(bounds_left, left)
                bounds_bottom = max(bounds_bottom, bottom)
                bounds_right = min(bounds_right, right)
                bounds_top = min(bounds_top, top)

        # Align the cells
        bounds_transform, bounds_width, bounds_height = align_bounds(bounds_left,
                                                                     bounds_bottom,
                                                                     bounds_right,
                                                                     bounds_top,
                                                                     dst_res)

    else:

        bounds_width = int((bounds_right - bounds_left) / abs(dst_res[0]))
        bounds_height = int((bounds_top - bounds_bottom) / abs(dst_res[1]))

        bounds_transform = from_bounds(bounds_left,
                                       bounds_bottom,
                                       bounds_right,
                                       bounds_top,
                                       bounds_width,
                                       bounds_height)

    if return_bounds:
        return array_bounds(bounds_height, bounds_width, bounds_transform)
    else:
        return bounds_transform, bounds_width, bounds_height


def warp_images(filenames,
                bounds_by='union',
                bounds=None,
                crs=None,
                res=None,
                nodata=0,
                resampling='nearest',
                warp_mem_limit=512,
                num_threads=1,
                tac=None):

    """
    Transforms a list of images to a common grid

    Args:
        filenames (list): The file names to mosaic.
        bounds_by (Optional[str]): How to concatenate the output extent. Choices are ['intersection', 'union'].
        bounds (Optional[tuple]): The extent bounds to warp to. If not give, the union of all images is used.
        crs (Optional[object]): The CRS to warp to.
        res (Optional[tuple]): The cell resolution to warp to.
        nodata (Optional[int or float]): The 'no data' value.
        resampling (Optional[str]): The resampling method. Choices are ['average', 'bilinear', 'cubic',
            'cubic_spline', 'gauss', 'lanczos', 'max', 'med', 'min', 'mode', 'nearest'].
        warp_mem_limit (Optional[int]): The memory limit (in MB) for the ``rasterio.vrt.WarpedVRT`` function.
        num_threads (Optional[int]): The number of warp worker threads.
        tac (Optional[tuple]): Target aligned raster coordinates (x, y).

    Returns:
        ``list`` of ``rasterio.vrt.WarpedVRT`` objects
    """

    if resampling not in ['average', 'bilinear', 'cubic', 'cubic_spline',
                          'gauss', 'lanczos', 'max', 'med', 'min', 'mode', 'nearest']:

        logger.warning("  The resampling method is not supported by rasterio. Setting to 'nearest'")

        resampling = 'nearest'

    warp_kwargs = {'resampling': resampling,
                   'crs': crs,
                   'res': res,
                   'nodata': nodata,
                   'warp_mem_limit': warp_mem_limit,
                   'num_threads': num_threads,
                   'tac': tac}

    if bounds:
        warp_kwargs['bounds'] = bounds
    else:

        # Get the union bounds of all images.
        #   *Target-aligned-pixels are returned.
        warp_kwargs['bounds'] = get_file_bounds(filenames,
                                                bounds_by=bounds_by,
                                                crs=crs,
                                                res=res,
                                                return_bounds=True)

    return [warp(fn, **warp_kwargs) for fn in filenames]


def get_ref_image_meta(filename):

    """
    Gets warping information from a reference image

    Args:
        filename (str): The file name to get information from.

    Returns:
        ``collections.namedtuple``
    """

    WarpInfo = namedtuple('WarpInfo', 'bounds crs res')

    with rio.open(filename) as src:

        bounds = src.bounds
        crs = src.crs
        res = src.res

    return WarpInfo(bounds=bounds, crs=crs, res=res)


def warp(filename,
         resampling='nearest',
         bounds=None,
         crs=None,
         res=None,
         nodata=0,
         warp_mem_limit=512,
         num_threads=1,
         tap=False,
         tac=None):

    """
    Warps an image to a VRT object

    Args:
        filename (str): The input file name.
        resampling (Optional[str]): The resampling method. Choices are ['average', 'bilinear', 'cubic',
            'cubic_spline', 'gauss', 'lanczos', 'max', 'med', 'min', 'mode', 'nearest'].
        bounds (Optional[tuple]): The extent bounds to warp to.
        crs (Optional[``CRS`` | int | dict | str]): The CRS to warp to.
        res (Optional[tuple]): The cell resolution to warp to.
        nodata (Optional[int or float]): The 'no data' value.
        warp_mem_limit (Optional[int]): The memory limit (in MB) for the ``rasterio.vrt.WarpedVRT`` function.
        num_threads (Optional[int]): The number of warp worker threads.
        tap (Optional[bool]): Whether to target align pixels.
        tac (Optional[tuple]): Target aligned raster coordinates (x, y).

    Returns:
        ``rasterio.vrt.WarpedVRT``
    """

    with rio.open(filename) as src:

        if res:
            dst_res = check_res(res)
        else:
            dst_res = src.res

        if crs:
            dst_crs = check_crs(crs)
        else:
            dst_crs = src.crs

        # Check if the data need to be subset
        if bounds and (bounds != src.bounds):

            if isinstance(bounds, str):

                if bounds.startswith('BoundingBox'):

                    bounds_str = bounds.replace('BoundingBox(', '').split(',')

                    for str_ in bounds_str:

                        if str_.strip().startswith('left='):
                            left_coord = float(str_.strip().split('=')[1].replace(')', ''))
                        elif str_.strip().startswith('bottom='):
                            bottom_coord = float(str_.strip().split('=')[1].replace(')', ''))
                        elif str_.strip().startswith('right='):
                            right_coord = float(str_.strip().split('=')[1].replace(')', ''))
                        elif str_.strip().startswith('top='):
                            top_coord = float(str_.strip().split('=')[1].replace(')', ''))

                    dst_bounds = BoundingBox(left=left_coord,
                                             bottom=bottom_coord,
                                             right=right_coord,
                                             top=top_coord)

                else:
                    logger.exception('  The bounds were not accepted.')

            else:

                dst_bounds = BoundingBox(left=bounds[0],
                                         bottom=bounds[1],
                                         right=bounds[2],
                                         top=bounds[3])

        else:
            dst_bounds = src.bounds

        dst_width = int((dst_bounds.right - dst_bounds.left) / dst_res[0])
        dst_height = int((dst_bounds.top - dst_bounds.bottom) / dst_res[1])

        # Do not warp if all the key metadata match the reference information
        if (src.bounds == bounds) and \
                (src.res == dst_res) and \
                (src.crs == dst_crs) and \
                (src.width == dst_width) and \
                (src.height == dst_height):

            output = filename

        else:

            dst_transform = Affine(dst_res[0], 0.0, dst_bounds.left, 0.0, -dst_res[1], dst_bounds.top)

            if tac:

                # Align the cells to target coordinates
                tap_left = tac[0][np.abs(tac[0] - dst_bounds.left).argmin()]
                tap_top = tac[1][np.abs(tac[1] - dst_bounds.top).argmin()]

                dst_transform = Affine(dst_res[0], 0.0, tap_left, 0.0, -dst_res[1], tap_top)

            if tap:

                # Align the cells to the resolution
                dst_transform, dst_width, dst_height = aligned_target(dst_transform,
                                                                      dst_width,
                                                                      dst_height,
                                                                      dst_res)

            vrt_options = {'resampling': getattr(Resampling, resampling),
                           'crs': dst_crs,
                           'transform': dst_transform,
                           'height': dst_height,
                           'width': dst_width,
                           'nodata': nodata,
                           'warp_mem_limit': warp_mem_limit,
                           'warp_extras': {'multi': True,
                                           'warp_option': 'NUM_THREADS={:d}'.format(num_threads)}}

            with WarpedVRT(src, **vrt_options) as vrt:
                output = vrt

    return output


def transform_crs(data_src,
                  dst_crs=None,
                  dst_res=None,
                  dst_width=None,
                  dst_height=None,
                  dst_bounds=None,
                  resampling='nearest',
                  warp_mem_limit=512,
                  num_threads=1):

    """
    Transforms a DataArray to a new coordinate reference system.

    Args:
        data_src (DataArray): The data to transform.
        dst_crs (Optional[CRS | int | dict | str]): The destination CRS.
        dst_res (Optional[float | int | tuple]): The destination resolution.
        dst_width (Optional[int]): The destination width. Cannot be used with ``dst_res``.
        dst_height (Optional[int]): The destination height. Cannot be used with ``dst_res``.
        dst_bounds (Optional[BoundingBox | tuple]): The destination bounds, as a ``rasterio.coords.BoundingBox``
            or as a tuple of (left, bottom, right, top).
        resampling (Optional[str]): The resampling method if ``filename`` is a ``list``.
            Choices are ['average', 'bilinear', 'cubic', 'cubic_spline', 'gauss', 'lanczos', 'max', 'med', 'min', 'mode', 'nearest'].
        warp_mem_limit (Optional[int]): The warp memory limit.
        num_threads (Optional[int]): The number of parallel threads.

    Returns:
        ``numpy.ndarray`` ``tuple`` ``CRS``
    """

    if dst_crs:
        dst_crs = check_crs(dst_crs)
    else:
        dst_crs = data_src.crs

    if isinstance(dst_res, int) or isinstance(dst_res, float) or isinstance(dst_res, tuple):

        dst_res = check_res(dst_res)

        dst_width = None
        dst_height = None

    else:

        if not isinstance(dst_width, int):
            if not isinstance(dst_height, int):
                dst_res = data_src.res

    if not dst_res:

        if not dst_width:
            dst_width = data_src.gw.ncols

        if not dst_height:
            dst_height = data_src.gw.nrows

    if not dst_bounds:
        dst_bounds = data_src.gw.bounds

    if not isinstance(dst_bounds, BoundingBox):

        dst_bounds = BoundingBox(left=dst_bounds[0],
                                 bottom=dst_bounds[1],
                                 right=dst_bounds[2],
                                 top=dst_bounds[3])

    dst_transform, dst_width, dst_height = calculate_default_transform(data_src.crs,
                                                                       dst_crs,
                                                                       data_src.gw.ncols,
                                                                       data_src.gw.nrows,
                                                                       left=dst_bounds.left,
                                                                       bottom=dst_bounds.bottom,
                                                                       right=dst_bounds.right,
                                                                       top=dst_bounds.top,
                                                                       dst_width=dst_width,
                                                                       dst_height=dst_height,
                                                                       resolution=dst_res)

    if not dst_res:

        cellx = (dst_bounds.right - dst_bounds.left) / dst_width
        celly = (dst_bounds.top - dst_bounds.bottom) / dst_height

        dst_res = (cellx, celly)

    transformed_array = list()

    for band in range(0, data_src.gw.nbands):

        destination = np.zeros((dst_height,
                                dst_width), dtype=data_src.dtype)

        data_dst, dst_transform = reproject(data_src[band, :, :].data.compute(num_workers=num_threads),
                                            destination,
                                            src_transform=data_src.transform,
                                            src_crs=data_src.crs,
                                            dst_transform=dst_transform,
                                            dst_crs=dst_crs,
                                            resampling=getattr(Resampling, resampling),
                                            dst_resolution=dst_res,
                                            warp_mem_limit=warp_mem_limit,
                                            num_threads=num_threads)

        transformed_array.append(data_dst)

    return np.array(transformed_array), dst_transform, dst_crs



import os
import logging


_FORMAT = '%(asctime)s:%(levelname)s:%(lineno)s:%(module)s.%(funcName)s:%(message)s'
_formatter = logging.Formatter(_FORMAT, '%H:%M:%S')
_handler = logging.StreamHandler()
_handler.setFormatter(_formatter)

logging.basicConfig(filename=os.path.join(os.path.abspath(os.path.dirname(__file__)), 'geowombat.log'),
                    filemode='w',
                    level=logging.DEBUG)

logger = logging.getLogger(__name__)
logger.addHandler(_handler)
logger.setLevel(logging.INFO)



from dask.distributed import Client, LocalCluster


class Cluster(object):

    """
    Wrapper for ``dask`` clients

    Best practices:

       "By "node" people typically mean a physical or virtual machine. That node can run several programs or processes
        at once (much like how my computer can run a web browser and text editor at once). Each process can parallelize
        within itself with many threads. Processes have isolated memory environments, meaning that sharing data within
        a process is free, while sharing data between processes is expensive.

        Typically things work best on larger nodes (like 36 cores) if you cut them up into a few processes, each of
        which have several threads. You want the number of processes times the number of threads to equal the number
        of cores. So for example you might do something like the following for a 36 core machine:

            Four processes with nine threads each
            Twelve processes with three threads each
            One process with thirty-six threads

        Typically one decides between these choices based on the workload. The difference here is due to Python's
        Global Interpreter Lock, which limits parallelism for some kinds of data. If you are working mostly with
        Numpy, Pandas, Scikit-Learn, or other numerical programming libraries in Python then you don't need to worry
        about the GIL, and you probably want to prefer few processes with many threads each. This helps because it
        allows data to move freely between your cores because it all lives in the same process. However, if you're
        doing mostly Pure Python programming, like dealing with text data, dictionaries/lists/sets, and doing most of
        your computation in tight Python for loops then you'll want to prefer having many processes with few threads
        each. This incurs extra communication costs, but lets you bypass the GIL.

        In short, if you're using mostly numpy/pandas-style data, try to get at least eight threads or so in a process.
        Otherwise, maybe go for only two threads in a process."

        --MRocklin (https://stackoverflow.com/questions/51099685/best-practices-in-setting-number-of-dask-workers)

    Examples:
        >>> # I/O-heavy task with 8 nodes
        >>> cluster = Cluster(n_workers=4,
        >>>                   threads_per_worker=2,
        >>>                   scheduler_port=0,
        >>>                   processes=False)
        >>>
        >>> # Task with little need of the GIL with 16 nodes
        >>> cluster = Cluster(n_workers=2,
        >>>                   threads_per_worker=8,
        >>>                   scheduler_port=0,
        >>>                   processes=False)
    """

    def __init__(self, **kwargs):

        self.kwargs = kwargs
        self.cluster = None
        self.client = None

    def start(self):

        self.cluster = LocalCluster(**self.kwargs)
        self.client = Client(self.cluster)

    def stop(self):

        self.client.close()
        self.cluster.close()

        self.client = None
        self.cluster = None



from copy import copy
from collections import namedtuple

from .angles import relative_azimuth
from ..core.util import project_coords
from ..errors import logger

import numpy as np
import numexpr as ne
import xarray as xr
import dask
import dask.array as da


def dtor(x):
    """Converts degrees to radians"""
    return x * np.pi / 180.0


def get_phase_angle_dask(cos_vza, cos_sza, sin_vza, sin_sza, cos_raa):

    """
    Calculates the Phase angle component
    """

    cos_phase_angle = cos_vza * cos_sza + sin_vza * sin_sza * cos_raa

    cos_phase_angle = da.clip(cos_phase_angle, -1, 1)

    phase_angle = da.arccos(cos_phase_angle)
    sin_phase_angle = da.sin(phase_angle)

    return cos_phase_angle, phase_angle, sin_phase_angle


def __ross_kernel_part_delayed(angle_info, global_args):

    """
    Calculates the main part of Ross kernel
    """

    RossKernelOutputs = namedtuple('RossKernelOutputs', 'cos1 cos2 sin1 sin2 cos3 rosselement cosphaang phaang sinphaang ross')

    __cos1 = da.cos(da.from_delayed(angle_info.vza, shape=(global_args.size,), dtype=global_args.dtype).rechunk(global_args.flat_chunks))
    __cos2 = da.cos(da.from_delayed(angle_info.sza, shape=(global_args.size,), dtype=global_args.dtype).rechunk(global_args.flat_chunks))

    __sin1 = da.sin(da.from_delayed(angle_info.vza, shape=(global_args.size,), dtype=global_args.dtype).rechunk(global_args.flat_chunks))
    __sin2 = da.sin(da.from_delayed(angle_info.sza, shape=(global_args.size,), dtype=global_args.dtype).rechunk(global_args.flat_chunks))
    __cos3 = da.cos(da.from_delayed(angle_info.raa, shape=(global_args.size,), dtype=global_args.dtype).rechunk(global_args.flat_chunks))

    __cosphaang, __phaang, __sinphaang = get_phase_angle_dask(__cos1, __cos2, __sin1, __sin2, __cos3)

    rosselement = (global_args.m_pi2 - __phaang) * __cosphaang + __sinphaang

    return RossKernelOutputs(cos1=__cos1,
                             cos2=__cos2,
                             sin1=__sin1,
                             sin2=__sin2,
                             cos3=__cos3,
                             rosselement=rosselement,
                             cosphaang=__cosphaang,
                             phaang=__phaang,
                             sinphaang=__sinphaang,
                             ross=None)


def __ross_kernel_part_delayed_2d(angle_info, global_args):

    """
    Calculates the main part of Ross kernel
    """

    RossKernelOutputs = namedtuple('RossKernelOutputs', 'cos_vza cos_sza sin_vza sin_sza cos_raa ross_element cos_phase_angle phase_angle sin_phase_angle ross')

    cos_vza = da.cos(angle_info.vza_rad)
    cos_sza = da.cos(angle_info.sza_rad)

    sin_vza = da.sin(angle_info.vza_rad)
    sin_sza = da.sin(angle_info.sza_rad)
    cos_raa = da.cos(angle_info.raa_rad)

    cos_phase_angle, phase_angle, sin_phase_angle = get_phase_angle_dask(cos_vza, cos_sza, sin_vza, sin_sza, cos_raa)

    ross_element = (global_args.m_pi2 - phase_angle) * cos_phase_angle + sin_phase_angle

    return RossKernelOutputs(cos_vza=cos_vza,
                             cos_sza=cos_sza,
                             sin_vza=sin_vza,
                             sin_sza=sin_sza,
                             cos_raa=cos_raa,
                             ross_element=ross_element,
                             cos_phase_angle=cos_phase_angle,
                             phase_angle=phase_angle,
                             sin_phase_angle=sin_phase_angle,
                             ross=None)


def ross_thin_delayed(angle_info, global_args):

    """
    Public method - call to calculate Ross Thin kernel
    """

    RossThinOutputs = namedtuple('RossThinOutputs', 'ross phase_angle')

    ross_kernel_outputs = __ross_kernel_part_delayed_2d(angle_info, global_args)

    ross_ = ross_kernel_outputs.ross_element / (ross_kernel_outputs.cos_vza * ross_kernel_outputs.cos_sza)

    return RossThinOutputs(ross=ross_, phase_angle=ross_kernel_outputs.phase_angle)


def ross_thick_delayed(angle_info, global_args):

    """
    Public method - call to calculate RossThick kernel
    """

    RossThickOutputs = namedtuple('RossThickOutputs', 'ross phase_angle')

    # RossKernelOutputs = namedtuple('RossKernelOutputs', 'cos1 cos2 sin1 sin2 cos3 rosselement cosphaang phaang sinphaang ross')
    ross_kernel_outputs = __ross_kernel_part_delayed_2d(angle_info, global_args)

    ross_ = ross_kernel_outputs.ross_element / (ross_kernel_outputs.cos_vza + ross_kernel_outputs.cos_sza)

    return RossThickOutputs(ross=ross_, phase_angle=ross_kernel_outputs.phase_angle)


def ross_kernel_dask(angle_info, global_args):

    """
    Public method - call to calculate Ross Kernel
    """

    # RossKernelOutputs = namedtuple('RossKernelOutputs', 'cos1 cos2 sin1 sin2 cos3 rosselement cosphaang phaang sinphaang ross')
    if global_args.ross_type.lower() == 'thin':
        ross_kernel_outputs = ross_thin_delayed(angle_info, global_args)
    else:
        ross_kernel_outputs = ross_thick_delayed(angle_info, global_args)

    if global_args.ross_hs:
        ross = ross_kernel_outputs.ross * (1.0 + 1.0 / (1.0 + ross_kernel_outputs.phase_angle / 0.25))
    else:
        ross = ross_kernel_outputs.ross

    return ross


def get_pangles_dask(tan1, global_args):

    """
    Applies B/R transformation for ellipse shape
    """

    t = global_args.br * tan1

    t = da.where(t < 0, 0, t)

    # w = np.where(t < 0.)[0]
    # t[w] = 0.0

    angp = da.arctan(t)
    s = da.sin(angp)
    c = da.cos(angp)

    # have to make sure c is not 0
    c = da.where(c == 0, global_args.nearly_zero, c)

    # w = np.where(c == 0)[0]
    # c[w] = global_args.nearly_zero

    return c, s, t


def get_distance_dask(__tan1, __tan2, __cos3):

    """
    Gets distance component of Li kernels
    """

    temp = __tan1 * __tan1 + __tan2 * __tan2 - 2.0 * __tan1 * __tan2 * __cos3

    temp = da.where(temp < 0, 0, temp)

    # w = np.where(temp < 0)[0]
    # temp[w] = 0.0

    # TODO
    # self.__temp = temp  # used by other functions ??

    return da.sqrt(temp)


def get_overlap_dask(__cos1, __cos2, __tan1, __tan2, __sin3, __distance, global_args):

    """
    Applies HB ratio transformation
    """

    OverlapInfo = namedtuple('OverlapInfo', 'tvar sint overlap temp')

    __temp = (1.0 / __cos1) + (1.0 / __cos2)

    __cost = global_args.hb * da.sqrt(__distance * __distance + __tan1 * __tan1 * __tan2 * __tan2 * __sin3 * __sin3) / __temp

    __cost = da.clip(__cost, -1, 1)

    __tvar = da.arccos(__cost)
    __sint = da.sin(__tvar)
    __overlap = global_args.m_1_pi * (__tvar - __sint * __cost) * __temp

    __overlap = da.where(__overlap < 0, 0, __overlap)

    return OverlapInfo(tvar=__tvar, sint=__sint, overlap=__overlap, temp=__temp)


def li_kernel_delayed(angle_info, global_args):

    """Private method - call to calculate Li Kernel"""

    # LiKernelOutputs = namedtuple('LiKernelOutputs', 'li phi cos1 cos2 cos3 sin1 sin2 sin3 tan1 tan2 tanti tantv cosphaang phaang sinphaang distance')

    # at some point add in LiGround kernel & LiTransit
    # TODO
    # if self.LiType == 'Roujean':
    #     return self.RoujeanKernel()

    # first make sure its in range 0 to 2 pi
    __phi = da.fabs((da.from_delayed(angle_info.raa, shape=(global_args.size,), dtype=global_args.dtype).rechunk(global_args.flat_chunks) % (2.0 * global_args.m_pi)))
    __cos3 = da.cos(__phi)
    __sin3 = da.sin(__phi)
    __tanti = da.tan(da.from_delayed(angle_info.sza, shape=(global_args.size,), dtype=global_args.dtype).rechunk(global_args.flat_chunks))
    __tantv = da.tan(da.from_delayed(angle_info.vza, shape=(global_args.size,), dtype=global_args.dtype).rechunk(global_args.flat_chunks))
    __cos1, __sin1, __tan1 = get_pangles_dask(__tantv, global_args)
    __cos2, __sin2, __tan2 = get_pangles_dask(__tanti, global_args)

    # sets cos & sin phase angle terms
    __cosphaang, __phaang, __sinphaang = get_phase_angle_dask(__cos1, __cos2, __sin1, __sin2, __cos3)

    __distance = get_distance_dask(__tan1, __tan2, __cos3)

    # OverlapInfo = namedtuple('OverlapInfo', 'tvar sint overlap temp')
    overlap_info = get_overlap_dask(__cos1, __cos2, __tan1, __tan2, __sin3, __distance, global_args)

    if global_args.li_type.lower() == 'sparse':

        if global_args.recip_flag == True:
            li = overlap_info.overlap - overlap_info.temp + 0.5 * (1.0 + __cosphaang) / __cos1 / __cos2
        else:
            li = overlap_info.overlap - overlap_info.temp + 0.5 * (1.0 + __cosphaang) / __cos1

    # TODO
    # else:
    #     if self.LiType == 'Dense':
    #         if self.RecipFlag:
    #             self.Li = (1.0 + self.__cosphaang) / (
    #             self.__cos1 * self.__cos2 * (self.__temp - self.__overlap)) - 2.0;
    #         else:
    #             self.Li = (1.0 + self.__cosphaang) / (self.__cos1 * (self.__temp - self.__overlap)) - 2.0;
    #     else:
    #         B = self.__temp - self.__overlap
    #         w = np.where(B <= 2.0)
    #         self.Li = B * 0.0
    #         if self.RecipFlag == True:
    #             Li = self.__overlap - self.__temp + 0.5 * (1. + self.__cosphaang) / self.__cos1 / self.__cos2;
    #         else:
    #             Li = self.__overlap - self.__temp + 0.5 * (1. + self.__cosphaang) / self.__cos1;
    #         self.Li[w] = Li[w]
    #
    #         w = np.where(B > 2.0)
    #         if self.RecipFlag:
    #             Li = (1.0 + self.__cosphaang) / (self.__cos1 * self.__cos2 * (self.__temp - self.__overlap)) - 2.0;
    #         else:
    #             Li = (1.0 + self.__cosphaang) / (self.__cos1 * (self.__temp - self.__overlap)) - 2.0;
    #         self.Li[w] = Li[w]

    # return LiKernelOutputs(li=li, phi=__phi, cos1=__cos1, cos2=__cos2, cos3=__cos3, sin1=__sin1, sin2=__sin2, sin3=__sin3,
    #                        tan1=__tan1, tan2=__tan2, tanti=__tanti,
    #                        tantv=__tantv, cosphaang=__cosphaang, phaang=__phaang, sinphaang=__sinphaang,
    #                        distance=__distance)

    return li


def li_kernel_dask(angle_info, global_args):

    """
    Private method - call to calculate Li Kernel
    """

    # first make sure its in range 0 to 2 pi
    __phi = da.fabs((angle_info.raa_rad % (2.0 * global_args.m_pi)))
    __cos3 = da.cos(__phi)
    __sin3 = da.sin(__phi)

    __tanti = da.tan(angle_info.sza_rad)
    __tantv = da.tan(angle_info.vza_rad)

    __cos1, __sin1, __tan1 = get_pangles_dask(__tantv, global_args)
    __cos2, __sin2, __tan2 = get_pangles_dask(__tanti, global_args)

    # sets cos & sin phase angle terms
    __cosphaang, __phaang, __sinphaang = get_phase_angle_dask(__cos1, __cos2, __sin1, __sin2, __cos3)

    __distance = get_distance_dask(__tan1, __tan2, __cos3)

    # OverlapInfo = namedtuple('OverlapInfo', 'tvar sint overlap temp')
    overlap_info = get_overlap_dask(__cos1, __cos2, __tan1, __tan2, __sin3, __distance, global_args)

    if global_args.li_type.lower() == 'sparse':

        if global_args.recip_flag:
            li = overlap_info.overlap - overlap_info.temp + 0.5 * (1.0 + __cosphaang) / __cos1 / __cos2
        else:
            li = overlap_info.overlap - overlap_info.temp + 0.5 * (1.0 + __cosphaang) / __cos1

    else:

        if global_args.li_type.lower() == 'dense':

            if global_args.recip_flag:
                li = (1.0 + __cosphaang) / (__cos1 * __cos2 * (overlap_info.temp - overlap_info.overlap)) - 2.0
            else:
                li = (1.0 + __cosphaang) / (__cos1 * (overlap_info.temp - overlap_info.overlap)) - 2.0

        else:

            b = overlap_info.temp - overlap_info.overlap
            li = b * 0.0

            if global_args.recip_flag:
                li_ = global_args.overlap - global_args.temp + 0.5 * (1.0 + __cosphaang) / __cos1 / __cos2
            else:
                li_ = overlap_info.overlap - overlap_info.temp + 0.5 * (1.0 + __cosphaang) / __cos1

            li = da.where(b <= 2, li_, li)

            if global_args.recip_flag:
                li_ = (1.0 + __cosphaang) / (__cos1 * __cos2 * (overlap_info.temp - overlap_info.overlap)) - 2.0
            else:
                li_ = (1.0 + __cosphaang) / (__cos1 * (overlap_info.temp - overlap_info.overlap)) - 2.0

            li = da.where(b <= 2, li_, li)

    return li


@dask.delayed
def set_angle_info_delayed(vza, sza, raa, global_args):

    """
    Store and organizes the input angle data
    """

    AngleInfo = namedtuple('AngleInfo', 'n vza_degrees sza_degrees raa_degrees vza sza raa')

    vza_degrees = np.array(vza).flatten()
    sza_degrees = np.array(sza).flatten()
    raa_degrees = np.array(raa).flatten()

    n = vza_degrees.shape[0]

    if n != len(sza_degrees) or n != len(raa_degrees):

        logger.error('kernels: inconsistent number of samples in vza, sza and raa data: ' + str(len(vza_degrees)) + ', ' + str(len(sza_degrees)) + ', ' + str(len(raa_degrees)))
        print(vza_degrees)
        print(sza_degrees)
        print(raa_degrees)
        return [-1]

    if global_args.normalize >= 1:

        # calculate nadir term by extending array
        vza_degrees = np.array(vza_degrees.tolist() + [0.0]).flatten()
        sza_degrees = np.array(sza_degrees.tolist() + [global_args.nbar]).flatten()
        raa_degrees = np.array(raa_degrees.tolist() + [0.0]).flatten()

        n = len(vza_degrees)

    vza = dtor(vza_degrees)
    sza = dtor(sza_degrees)  # -1 to make HS direction for raa = 0
    raa = dtor(raa_degrees)

    w = np.where(vza < 0)[0]
    vza[w] = -vza[w]
    raa[w] = raa[w] + global_args.m_pi
    w = np.where(sza < 0)[0]
    sza[w] = -sza[w]
    raa[w] = raa[w] + global_args.m_pi

    return AngleInfo(n=n,
                     vza_degrees=vza_degrees,
                     sza_degrees=sza_degrees,
                     raa_degrees=raa_degrees,
                     vza=vza,
                     sza=sza,
                     raa=raa)


def set_angle_info_delayed_2d(vza, sza, raa, global_args):

    """
    Store and organizes the input angle data
    """

    AngleInfo = namedtuple('AngleInfo', 'vza sza raa vza_rad sza_rad raa_rad')

    # if global_args.normalize >= 1:
    #
    #     # calculate nadir term by extending array
    #     vza = np.array(vza.tolist() + [0.0]).flatten()
    #     sza = np.array(sza.tolist() + [global_args.nbar]).flatten()
    #     raa = np.array(raa.tolist() + [0.0]).flatten()
    #
    #     n = len(vza_degrees)

    vza_rad = da.deg2rad(vza)
    sza_rad = da.deg2rad(sza)
    raa_rad = da.deg2rad(raa)

    vza_abs = da.fabs(vza_rad)
    sza_abs = da.fabs(sza_rad)

    raa_abs = da.where((vza_rad < 0) | (sza_rad < 0), global_args.m_pi, raa_rad)

    return AngleInfo(vza=vza,
                     sza=sza,
                     raa=raa,
                     vza_rad=vza_abs,
                     sza_rad=sza_abs,
                     raa_rad=raa_abs)


def post_process_ross_delayed(ross, angle_info, global_args):

    # li_norm = 0.0
    # ross_norm = 0.0
    # isotropic_norm = 0.0

    # if we are normalising the last element of self.Isotropic, self.Ross and self.Li contain the nadir-nadir kernel

    if global_args.normalize >= 1:

        # normalise nbar-nadir (so kernel is 0 at nbar-nadir)
        ross_norm = ross[-1]
        ross = ross - ross_norm

        # depreciate length of arrays
        # ross = ross[0:-1]

        # TODO
        # if hasattr(self, 'Isotropic'):
        #     self.Isotropic = self.Isotropic[0:-1]

        # angle_info = angle_info._update(vza_degrees=angle_info.vza_degrees[0:-1])
        # angle_info = angle_info._update(sza_degrees=angle_info.sza_degrees[0:-1])
        # angle_info = angle_info._update(raa_degrees=angle_info.raa_degrees[0:-1])

        # angle_info = angle_info._update(n=len(angle_info.vza_degrees))
        # angle_info = angle_info._update(vza=angle_info.vza[0:-1])
        # angle_info = angle_info._update(sza=angle_info.sza[0:-1])
        # angle_info = angle_info._update(raa=angle_info.raa[0:-1])

        # return angle_info, ross_kernel_outputs, li_kernel_outputs

    return ross


def post_process_li_delayed(li, angle_info, global_args):

    """
    Handles with normalisation
    """

    # li_norm = 0.0
    # ross_norm = 0.0
    # isotropic_norm = 0.0

    # if we are normalising the last element of self.Isotropic, self.Ross and self.Li contain the nadir-nadir kernel

    if global_args.normalize >= 1:

        # normalise nbar-nadir (so kernel is 0 at nbar-nadir)
        li_norm = li[-1]
        li = li - li_norm

        # depreciate length of arrays
        # li = li[0:-1]

        # TODO
        # if hasattr(self, 'Isotropic'):
        #     self.Isotropic = self.Isotropic[0:-1]

        # angle_info = angle_info._update(vza_degrees=angle_info.vza_degrees[0:-1])
        # angle_info = angle_info._update(sza_degrees=angle_info.sza_degrees[0:-1])
        # angle_info = angle_info._update(raa_degrees=angle_info.raa_degrees[0:-1])

        # angle_info = angle_info._update(n=len(angle_info.vza_degrees))
        # angle_info = angle_info._update(vza=angle_info.vza[0:-1])
        # angle_info = angle_info._update(sza=angle_info.sza[0:-1])
        # angle_info = angle_info._update(raa=angle_info.raa[0:-1])

        # return angle_info, ross_kernel_outputs, li_kernel_outputs

    return li


@dask.delayed
def post_process_delayed(ross_kernel_outputs, li_kernel_outputs, angle_info, global_args):

    """
    Handles with normalisation
    """

    li_norm = 0.0
    ross_norm = 0.0
    isotropic_norm = 0.0

    # if we are normalising the last element of self.Isotropic, self.Ross and self.Li contain the nadir-nadir kernel

    if global_args.normalize >= 1:

        # normalise nbar-nadir (so kernel is 0 at nbar-nadir)
        ross_norm = ross_kernel_outputs.ross[-1]
        li_norm = li_kernel_outputs.li[-1]
        ross_kernel_outputs = ross_kernel_outputs._update(ross=ross_kernel_outputs.ross - ross_norm)
        li_kernel_outputs = li_kernel_outputs._update(li=li_kernel_outputs.li - li_norm)

        # depreciate length of arrays
        ross_kernel_outputs = ross_kernel_outputs._update(ross=ross_kernel_outputs.ross[0:-1])
        li_kernel_outputs = li_kernel_outputs._update(li=li_kernel_outputs.li[0:-1])

        # TODO
        # if hasattr(self, 'Isotropic'):
        #     self.Isotropic = self.Isotropic[0:-1]

        angle_info = angle_info._update(vza_degrees=angle_info.vza_degrees[0:-1])
        angle_info = angle_info._update(sza_degrees=angle_info.sza_degrees[0:-1])
        angle_info = angle_info._update(raa_degrees=angle_info.raa_degrees[0:-1])

        angle_info = angle_info._update(n=len(angle_info.vza_degrees))
        angle_info = angle_info._update(vza=angle_info.vza[0:-1])
        angle_info = angle_info._update(sza=angle_info.sza[0:-1])
        angle_info = angle_info._update(raa=angle_info.raa[0:-1])

        return angle_info, ross_kernel_outputs, li_kernel_outputs


class Kernels(object):

    """
    Linear kernel models

    Source:
        Code taken from:
            https://github.com/jgomezdans/eoldas/blob/master/eoldas/kernels.py
    """

    def __init__(self,
                 vza,
                 sza,
                 raa,
                 delayed=False,
                 critical=1,
                 RossHS=True,
                 RecipFlag=True,
                 HB=2.0,
                 BR=1.0,
                 MODISSPARSE=True,
                 MODISDENSE=False,
                 RossType='Thick',
                 normalise=1,
                 normalize=0,
                 LiType='sparse',
                 doIntegrals=True,
                 BSAangles=None,
                 nbar=0.0):

        """
        The class creator sets up the kernels for some angle set. Default Li is MODISSPARSE parameter set

        The kernels are accessible from:
            self.Isotropic
            self.Ross
            self.Li
        The angles are accesible from:
            self.vza (or self.vzaDegrees)
            self.sza (or self.szaDegrees)
            self.raa (or self.raaDegrees)
            N.B. Hot spot direction is vza == sza and raa = 0.0
        Kernels integrals are acessible from:
            self.BSAangles (angles in degrees)
            self.BSA_Isotropic (directional-hemispherical integral of self.Isotropic)
            self.BSA_Ross (directional-hemispherical integral of self.Ross)
            self.BSA_Li (directional-hemispherical integral of self.Li)
            self.WSA_Isotropic (bi-hemispherical integral of self.Isotropic)
            self.WSA_Ross (bi-hemispherical integral of self.Ross)
            self.WSA_Li (bi-hemispherical integral of self.Li)
            N.B. You need to set the doIntegrals flag to True on creating an instance of the kernels class if you
            want access to integrals. The processing takes a bit of time.
        Printing methods are available:
            self.printIntegrals(header=True,reflectance=False)
            self.printKernels(header=True,reflectance=False)

        Required parameters:

            @param vza: an array containing view zenith angles in degrees
            @param sza: an array containing solar zenith angles in degrees
            @param raa: an array containing relative azimuth angles in degrees

        Options:
            @option critical=1: set to 1 to exit on error, 0 not to
            @option RecipFlag=True: Li reciprocal flag
            @option HB: Li kernel parameter HB
            @option BR: Li kernel parameter
            @option MODISSPARSE: set to True for default MODIS Li Sparse parameters (overrides BR and HB to 2.0 and 1.0)
            @option MODISDENSE: set to True for default MODIS Li Dense parameters (override BR and HB to 2.0 and 2.5)
            @option RossType: set to 'Thin' for Ross Thin (default) else 'Thick'
            @option LiType: set to 'Sparse' for LiSparse (default). Other options: 'Roujean', 'Dense'
            @option normalise: set to 1 to make kernels 0 at nadir view illumination (default), set to 0 for no normalisation (can also use US spelling, i.e. normalize)
            @option doIntegrals: set to True to calculate integrals of kernels numerically. Set to False not to calculate them. At some point will have Approx flag here as well.
            @option BSAangles: solar zenith angles at which to calculate directional-hemispherical integral of kernels (default 0-89 in steps of 1 degree). Units: degrees.
            @option nbar: the sza at which the isotropic term is set to if normalise=1 is turned on (default 0)

        Notes:
        Requires numpy. If you do integrals, this also requires scipy (or rather scipy.integrate)
        If you want to mimic the results in Wanner et al. 1995, I've set a special function called self.mimic at the end here.
        """

        GlobalArgs = namedtuple('GlobalArgs', 'nrows ncols size dtype flat_chunks nbar nearly_zero critical file_ outputfile li_type ross_type ross_hs do_integrals hb br normalize recip_flag m_pi m_pi2 m_pi4 m_1_pi')

        self.__setup(critical=critical,
                     RecipFlag=RecipFlag,
                     RossHS=RossHS,
                     HB=HB,
                     BR=BR,
                     MODISSPARSE=MODISSPARSE,
                     MODISDENSE=MODISDENSE,
                     RossType=RossType,
                     normalise=normalise,
                     normalize=normalize,
                     LiType=LiType,
                     doIntegrals=doIntegrals,
                     BSAangles=BSAangles,
                     nbar=nbar)

        if delayed:

            # TODO: make `flat_chunks` a user argument
            global_args = GlobalArgs(nrows=vza.shape[-2], ncols=vza.shape[-1], size=vza.size, dtype=vza.dtype.name,
                                     flat_chunks=np.array(list(vza.chunksize)).sum(),
                                     nbar=nbar, nearly_zero=self.__NEARLYZERO, critical=self.critical,
                                     file_=self.FILE, outputfile=self.outputFile,
                                     li_type=self.LiType, ross_type=self.RossType, ross_hs=self.RossHS,
                                     do_integrals=self.doIntegrals, hb=self.HB, br=self.BR, normalize=self.normalise,
                                     recip_flag=self.RecipFlag,
                                     m_pi=self.__M_PI, m_pi2=self.__M_PI_2, m_pi4=self.__M_PI_4, m_1_pi=self.__M_1_PI)

            # AngleInfo = namedtuple('AngleInfo', 'n vza sza raa vza_rad sza_rad raa_rad')
            # angle_info = set_angle_info_delayed(vza, sza, raa, global_args)
            angle_info = set_angle_info_delayed_2d(vza, sza, raa, global_args)

            # RossKernelOutputs = namedtuple('RossKernelOutputs', 'cos1 cos2 sin1 sin2 cos3 rosselement cosphaang phaang sinphaang ross')
            ross = ross_kernel_dask(angle_info, global_args)

            # LiKernelOutputs = namedtuple('LiKernelOutputs','li phi cos1 cos2 cos3 sin1 sin2 sin3 tan1 tan2 tanti tantv cosphaang phaang sinphaang distance')
            li = li_kernel_dask(angle_info, global_args)

            # TODO: add normalization for 2d dask arrays
            # ross = post_process_ross_delayed(ross, angle_info, global_args)
            # li = post_process_li_delayed(li, angle_info, global_args)

            # angle_info, ross_kernel_outputs, li_kernel_outputs = post_process_delayed(ross_kernel_outputs,
            #                                                                           li_kernel_outputs,
            #                                                                           angle_info,
            #                                                                           global_args)

            self.Ross = ross
            self.Li = li

        else:

            self.set_angle_info(vza, sza, raa)
            self.__doKernels()
            self.__postProcess()

    def __setup(self,
                critical=1,
                RecipFlag=True,
                RossHS=True,
                HB=2.0,
                BR=1.0,
                MODISSPARSE=True,
                MODISDENSE=False,
                RossType='Thick',
                normalise=1,
                normalize=0,
                LiType='Sparse',
                doIntegrals=True,
                BSAangles=None,
                nbar=0.0):

        self.nbar = nbar
        self.__NEARLYZERO = 1e-20
        self.critical = critical
        self.FILE = -1
        self.outputFile = ''
        # kernel options etc.
        self.LiType = LiType
        self.RossHS = RossHS
        self.doIntegrals = doIntegrals

        if MODISDENSE:

            LiType = 'Dense'
            self.HB = 2.
            self.BR = 2.5

        else:

            if MODISSPARSE:

                LiType = 'Sparse'
                self.HB = 2.0
                self.BR = 1.0

            else:

                self.HB = HB
                self.BR = BR

        # self.LiType = LiType
        self.RossType = RossType
        self.normalise = normalise
        self.RecipFlag = RecipFlag

        # some useful numbers
        self.__M_PI = np.pi
        self.__M_PI_2 = self.__M_PI * 0.5
        self.__M_PI_4 = self.__M_PI * 0.25
        self.__M_1_PI = 1.0 / self.__M_PI

        self.normalize = self.normalise

        # TODO
        # if self.doIntegrals:
        #     self.__integrateKernels(BSAangles=BSAangles)

        if (normalise >= 1) or (normalize >= 1):
            self.normalise = max(normalise, normalize)

    def __postProcess(self):

        """Private method for dealing with normalisation"""

        self.LiNorm = 0.
        self.RossNorm = 0.
        self.IsotropicNorm = 0.
        # if we are normalising the last element of self.Isotropic, self.Ross and self.Li  contain the nadir-nadir kernel

        if self.normalise >= 1:

            # normalise nbar-nadir (so kernel is 0 at nbar-nadir)
            self.RossNorm = self.Ross[-1]
            self.LiNorm = self.Li[-1]
            self.Ross = self.Ross - self.RossNorm
            self.Li = self.Li - self.LiNorm
            # depreciate length of arrays (well, teh ones we'll use again in any case)
            self.Ross = self.Ross[0:-1]
            self.Li = self.Li[0:-1]

            if hasattr(self, 'Isotropic'):
                self.Isotropic = self.Isotropic[0:-1]

            self.vzaDegrees = self.vzaDegrees[0:-1]
            self.szaDegrees = self.szaDegrees[0:-1]
            self.raaDegrees = self.raaDegrees[0:-1]
            self.N = len(self.vzaDegrees)
            self.vza = self.vza[0:-1]
            self.sza = self.sza[0:-1]
            self.raa = self.raa[0:-1]

    def __doKernels(self):

        """Private method to run the various kernel methods"""

        # the kernels
        # self.IsotropicKernel()
        self.RossKernel()
        self.LiKernel()

    def set_angle_info(self, vza, sza, raa):

        """Private method to store and organise the input angle data"""

        self.vzaDegrees = np.array([vza]).flatten()
        self.szaDegrees = np.array([sza]).flatten()
        self.raaDegrees = np.array([raa]).flatten()
        self.N = len(self.vzaDegrees)

        if (self.N != len(self.szaDegrees) or self.N != len(self.raaDegrees)):
            self.error('kernels: inconsistent number of samples in vza, sza and raa data: ' + str(
                len(self.vzaDegrees)) + ', ' + str(len(self.szaDegrees)) + ', ' + str(len(self.raaDegrees)),
                       critical=self.critical)
            print(self.vzaDegrees)
            print(self.szaDegrees)
            print(self.raaDegrees)
            return [-1]

        if (self.normalise >= 1):
            # calculate nadir term by extending array
            self.vzaDegrees = np.array(list(self.vzaDegrees) + [0.0]).flatten()
            self.szaDegrees = np.array(list(self.szaDegrees) + [self.nbar]).flatten()
            self.raaDegrees = np.array(list(self.raaDegrees) + [0.0]).flatten()
            # not N is one too many now
            self.N = len(self.vzaDegrees)

        self.vza = self.dtor(self.vzaDegrees)
        self.sza = self.dtor(self.szaDegrees)  # -1 to make HS direction for raa = 0
        self.raa = self.dtor(self.raaDegrees)
        w = np.where(self.vza < 0)[0]
        self.vza[w] = -self.vza[w]
        self.raa[w] = self.raa[w] + self.__M_PI
        w = np.where(self.sza < 0)[0]
        self.sza[w] = -self.sza[w]
        self.raa[w] = self.raa[w] + self.__M_PI

    def __integrateKernels(self, BSAangles=[]):

        """
        Private method to call integration functions for the kernels


             NB - this overwrites all kernel info ... so be careful how/where you call it
            @option: BSAangles=[] allows the user to set the sza angles at which directional-hemispherical intergal is
                calculated, else steps of 1 degree from 0 to 89 (though I wouldnt trust it down to 90)
            This function can be rather slow, so using fewer samples or an approximate function may be a god idea
        """

        import scipy.integrate
        if BSAangles == []:
            BSAangles = np.array(range(90)) * 1.0

        self.BSAangles = np.array(BSAangles).flatten()

        # isotropic integral
        self.BSA_Isotropic = np.zeros(len(self.BSAangles)) + 1.0
        self.BSA_Ross = np.zeros(len(self.BSAangles))
        self.BSA_Li = np.zeros(len(self.BSAangles))
        self.BSA_Isotropic_error = np.zeros(len(self.BSAangles))
        self.BSA_Ross_error = np.zeros(len(self.BSAangles))
        self.BSA_Li_error = np.zeros(len(self.BSAangles))

        i = 0
        mu = np.cos(self.BSAangles * self.__M_PI / 180.)
        for sza in self.BSAangles:
            # ross integral
            self.BSA_Ross[i], self.BSA_Ross_error[i] = scipy.integrate.dblquad(RossFunctionForIntegral, 0.0, 1.0,
                                                                               __gfun, __hfun, args=(sza, self))
            self.BSA_Li[i], self.BSA_Li_error[i] = scipy.integrate.dblquad(LiFunctionForIntegral, 0.0, 1.0, __gfun,
                                                                           __hfun, args=(sza, self))
            i = i + 1
        self.WSA_Ross = -2.0 * scipy.integrate.simps(self.BSA_Ross * mu, mu)
        self.WSA_Li = -2.0 * scipy.integrate.simps(self.BSA_Li * mu, mu)
        return

    def __GetPhaang(self):

        """Private method to calculate Phase angle component of kernel"""

        self.__cosphaang = self.__cos1 * self.__cos2 + self.__sin1 * self.__sin2 * self.__cos3
        # better check the bounds before arccos ... just to be safe
        w = np.where(self.__cosphaang < -1)[0]
        self.__cosphaang[w] = -1.0
        w = np.where(self.__cosphaang > 1)[0]
        self.__cosphaang[w] = 1.0
        self.__phaang = np.arccos(self.__cosphaang)
        self.__sinphaang = np.sin(self.__phaang)

    def __RossKernelPart(self):

        """Private method to calculate main part of Ross kernel"""

        self.__cos1 = np.cos(self.vza)
        self.__cos2 = np.cos(self.sza)

        self.__sin1 = np.sin(self.vza)
        self.__sin2 = np.sin(self.sza)
        self.__cos3 = np.cos(self.raa)
        self.__GetPhaang()

        self.rosselement = (self.__M_PI_2 - self.__phaang) * self.__cosphaang + self.__sinphaang

    def GetDistance(self):

        """Private method to get distance component of Li kernels"""

        temp = self.__tan1 * self.__tan1 + self.__tan2 * self.__tan2 - 2. * self.__tan1 * self.__tan2 * self.__cos3;
        w = np.where(temp < 0)[0]
        temp[w] = 0.0
        self.__temp = temp  # used by other functions ??
        distance = np.sqrt(temp)

        return distance

    def GetpAngles(self, tan1):

        """Private method to do B/R transformation for ellipse shape"""

        t = self.BR * tan1
        w = np.where(t < 0.)[0]
        t[w] = 0.0
        angp = np.arctan(t)
        s = np.sin(angp)
        c = np.cos(angp)
        # have to make sure c isnt 0
        w = np.where(c == 0)[0]
        c[w] = self.__NEARLYZERO

        return c, s, t

    def GetOverlap(self):

        """Private method to do HB ratio transformation"""

        self.__temp = 1. / self.__cos1 + 1. / self.__cos2

        self.__cost = self.HB * np.sqrt(
            self.__distance * self.__distance + self.__tan1 * self.__tan1 * self.__tan2 * self.__tan2 * self.__sin3 * self.__sin3) / self.__temp;
        w = np.where(self.__cost < -1)[0]
        self.__cost[w] = -1.0
        w = np.where(self.__cost > 1.0)[0]
        self.__cost[w] = 1.0
        self.__tvar = np.arccos(self.__cost)
        self.__sint = np.sin(self.__tvar)
        self.__overlap = self.__M_1_PI * (self.__tvar - self.__sint * self.__cost) * self.__temp
        w = np.where(self.__overlap < 0)[0]
        self.__overlap[w] = 0.0
        return

    def RoujeanKernel(self):

        """Private method - call to calculate Roujean shadowing kernel"""

        # first make sure its in range 0 to 2 pi
        self.__phi = np.abs((self.raa % (2. * self.__M_PI)))
        self.__cos3 = np.cos(self.__phi)
        self.__sin3 = np.sin(self.__phi)
        self.__tan1 = np.tan(self.sza)
        self.__tan2 = np.tan(self.vza)

        self.__distance = self.GetDistance()
        self.Li = 0.5 * self.__M_1_PI * (
        (self.__M_PI - self.__phi) * self.__cos3 + self.__sin3) * self.__tan1 * self.__tan2 - self.__M_1_PI * (
        self.__tan1 + self.__tan2 + self.__distance);
        return

    def LiKernel(self):

        """Private method - call to calculate Li Kernel"""

        # at some point add in LiGround kernel & LiTransit
        if self.LiType == 'Roujean':
            return self.RoujeanKernel()
        # first make sure its in range 0 to 2 pi
        self.__phi = np.abs((self.raa % (2. * self.__M_PI)))
        self.__cos3 = np.cos(self.__phi)
        self.__sin3 = np.sin(self.__phi)
        self.__tanti = np.tan(self.sza)
        self.__tantv = np.tan(self.vza)
        self.__cos1, self.__sin1, self.__tan1 = self.GetpAngles(self.__tantv);
        self.__cos2, self.__sin2, self.__tan2 = self.GetpAngles(self.__tanti);
        self.__GetPhaang();  # sets cos & sin phase angle terms
        self.__distance = self.GetDistance();  # sets self.temp
        self.GetOverlap();  # also sets self.temp
        if self.LiType == 'Sparse':
            if self.RecipFlag == True:
                self.Li = self.__overlap - self.__temp + 0.5 * (1. + self.__cosphaang) / self.__cos1 / self.__cos2;
            else:
                self.Li = self.__overlap - self.__temp + 0.5 * (1. + self.__cosphaang) / self.__cos1;
        else:
            if self.LiType == 'Dense':
                if self.RecipFlag:
                    self.Li = (1.0 + self.__cosphaang) / (
                    self.__cos1 * self.__cos2 * (self.__temp - self.__overlap)) - 2.0;
                else:
                    self.Li = (1.0 + self.__cosphaang) / (self.__cos1 * (self.__temp - self.__overlap)) - 2.0;
            else:
                B = self.__temp - self.__overlap
                w = np.where(B <= 2.0)
                self.Li = B * 0.0
                if self.RecipFlag == True:
                    Li = self.__overlap - self.__temp + 0.5 * (1. + self.__cosphaang) / self.__cos1 / self.__cos2;
                else:
                    Li = self.__overlap - self.__temp + 0.5 * (1. + self.__cosphaang) / self.__cos1;
                self.Li[w] = Li[w]

                w = np.where(B > 2.0)
                if self.RecipFlag:
                    Li = (1.0 + self.__cosphaang) / (self.__cos1 * self.__cos2 * (self.__temp - self.__overlap)) - 2.0;
                else:
                    Li = (1.0 + self.__cosphaang) / (self.__cos1 * (self.__temp - self.__overlap)) - 2.0;
                self.Li[w] = Li[w]
        return

    def IsotropicKernel(self):

        """Public method - call to calculate Isotropic kernel"""

        # default behaviour
        self.Isotropic = np.zeros(self.N) + 1.0

        return

    def RossThin(self):

        """Public method - call to calculate RossThin kernel"""

        self.__RossKernelPart()
        self.rosselement = self.rosselement / (self.__cos1 * self.__cos2)

        return

    def RossThick(self):

        """Public method - call to calculate RossThick kernel"""

        self.__RossKernelPart()
        self.rosselement = self.rosselement / (self.__cos1 + self.__cos2)

        return

    def RossKernel(self):

        """Public method - call to calculate Ross Kernel"""

        if self.RossType == 'Thin':
            self.RossThin()
        else:
            self.RossThick()

        self.Ross = self.rosselement

        if self.RossHS:
            self.Ross = self.Ross * (1. + 1. / (1. + self.__phaang / .25))

    def dtor(self, x):

        """Public method to convert degrees to radians"""

        return x * self.__M_PI / 180.

    def rtod(self, x):

        """Public method to convert radians to degrees"""

        return x * 180. / self.__M_PI

    def error(self, msg, critical=0, newline=1, code=-1):

        """
        Public method to do Class error reporting
        @param msg: error message
        @param critical: set to 1 if require exit (default critical=0)
        @param newline: set to 0 if newline not required (default newline=0)
        @param code: error code reported on exit if critical error (default code=-1)
        """

        if newline == 1:
            nl = '\n'
        else:
            nl = ''
        print(msg + nl)
        if critical == 1:
            logger.exception([code])

    def printIntegrals(self, header=True, reflectance=False):

        """
        Public method to print kernel integrals (to stdout only at present)
        """

        if (header == True):
            self.printer(
                '# ' + str(self.N) + ' samples Ross: ' + self.RossType + ' Li: ' + self.LiType + ' Reciprocal: ' + str(
                    self.RecipFlag) + ' normalisation: ' + str(self.normalise) + ' HB ' + str(self.HB) + ' BR ' + str(
                    self.BR) + '\n');
            self.printer('# WSA: Isotropic 1.0 Ross ' + str(self.WSA_Ross) + ' Li ' + str(self.WSA_Li))
            self.printer('# 1: SZA (degrees) 2: BSA Isotropic 3: BSA Ross 4: BSA Li')
            if (reflectance == True):
                self.printer(' ');
            self.printer('\n');

        for i in range(len(self.BSAangles)):
            self.printer(
                str(self.BSAangles[i]) + ' ' + str(self.BSA_Isotropic[i]) + ' ' + str(self.BSA_Ross[i]) + ' ' + str(
                    self.BSA_Li[i]))
            # print refl data if wanted
            if (reflectance == True):
                self.printer(' ');
            self.printer('\n');
        return

    def printKernels(self, header=True, reflectance=False, file=False):

        """
        Public method to print kernel values (to stdout only at present)
        """

        if (file != False):
            if (file != self.outputFile and self.FILE != -1):
                self.FILE.close()
            self.outputFile = file
            self.FILE = open(self.outputFile, 'w')

        if (header == True):
            self.printer(
                '# ' + str(self.N) + ' samples Ross: ' + self.RossType + ' Li: ' + self.LiType + ' Reciprocal: ' + str(
                    self.RecipFlag) + ' normalisation: ' + str(self.normalise) + ' HB ' + str(self.HB) + ' BR ' + str(
                    self.BR) + '\n');
            self.printer('# 1: VZA (degrees) 2: SZA (degrees) 3: RAA (degrees) 4: Isotropic 5: Ross 6: Li')
            if (reflectance == True):
                self.printer(' ');
            self.printer('\n');

        for i in range(self.N):
            self.printer(
                str(self.vzaDegrees[i]) + ' ' + str(self.szaDegrees[i]) + ' ' + str(self.raaDegrees[i]) + ' ' + str(
                    self.Isotropic[i]) + ' ' + str(self.Ross[i]) + ' ' + str(self.Li[i]))
            # print refl data if wanted
            if (reflectance == True):
                self.printer(' ');
            self.printer('\n');
        return

    def printer(self, msg):

        """
        Public print method ... make more flexible eg for printing to files at some point
        """

        if (self.FILE == -1):
            print(msg)
        else:
            self.FILE.write(msg)


def RossFunctionForIntegral(phi, mu, sza, self):
    # print phi
    # print mu
    # print sza
    # print '========'
    vza = np.arccos(mu)
    raa = self.rtod(phi)
    self.set_angle_info(vza, sza, raa)
    self.RossKernel()
    return mu * self.Ross[0] / np.pi


def LiFunctionForIntegral(phi, mu, sza, self):
    # print phi
    # print mu
    # print sza
    # print '========'
    vza = np.arccos(mu)
    raa = self.rtod(phi)
    self.set_angle_info(vza, sza, raa)
    self.LiKernel()
    return mu * self.Li[0] / np.pi


class RossLiKernels(object):

    def _get_kernels(self, central_latitude, solar_za, solar_az, sensor_za, sensor_az):

        # if isinstance(central_latitude, np.ndarray) or isinstance(central_latitude, xr.DataArray):
        #     delayed = True
        # else:
        #     delayed = False

        # Get the geometric scattering kernel.
        #
        # HLS uses a constant (per location) sun zenith angle (`solar_za`).
        # HLS uses 0 for sun azimuth angle (`solar_az`).
        # theta_v, theta_s, delta_gamma
        kl = Kernels(0.0, self.get_mean_sza(central_latitude), 0.0, delayed=False, doIntegrals=False)

        # Copy the geometric scattering
        #   coefficients so they are
        #   not overwritten by the
        #   volume scattering coefficients.
        self.geo_norm = copy(kl.Li)
        self.vol_norm = copy(kl.Ross)

        # Get the volume scattering kernel.
        #
        # theta_v=0 for nadir view zenith angle, theta_s, delta_gamma
        kl = Kernels(sensor_za.data,
                     solar_za.data,
                     relative_azimuth(solar_az, sensor_az).data,
                     delayed=True,
                     doIntegrals=False)

        self.geo_sensor = kl.Li
        self.vol_sensor = kl.Ross


class BRDF(RossLiKernels):

    """
    A class for Bidirectional Reflectance Distribution Function (BRDF) normalization
    """

    def __init__(self):

        self.geo_norm = None
        self.vol_norm = None
        self.geo_sensor = None
        self.vol_sensor = None

        # Setup the c-factor equation.
        #
        # `SA` = the sensor array
        self.c_equation = 'SA * ((fiso + fvol*vol_norm + fgeo*geo_norm) / (fiso + fvol*vol_sensor + fgeo*geo_sensor))'

        # A dictionary of BRDF kernel coefficients
        self.coeff_dict = dict(blue=dict(fiso=0.0774,
                                         fgeo=0.0079,
                                         fvol=0.0372),
                               green=dict(fiso=0.1306,
                                          fgeo=0.0178,
                                          fvol=0.058),
                               red=dict(fiso=0.169,
                                        fgeo=0.0227,
                                        fvol=0.0574),
                               nir=dict(fiso=0.3093,
                                        fgeo=0.033,
                                        fvol=0.1535),
                               swir1=dict(fiso=0.343,
                                          fgeo=0.0453,
                                          fvol=0.1154),
                               swir2=dict(fiso=0.2658,
                                          fgeo=0.0387,
                                          fvol=0.0639),
                               pan=dict(fiso=0.12567,
                                        fgeo=0.01613,
                                        fvol=0.0509))

    def _get_coeffs(self, sensor_band):
        return self.coeff_dict[sensor_band]

    @staticmethod
    def get_mean_sza(central_latitude):

        """
        Returns the mean solar zenith angle (SZA) as a function of the central latitude

        Args:
            central_latitude (float): The central latitude.

        Reference:

            See :cite:`zhang_etal_2016`

        Returns:
            ``float``
        """

        return 31.0076 + \
               -0.1272 * central_latitude + \
               0.01187 * (central_latitude ** 2) + \
               2.40e-05 * (central_latitude ** 3) + \
               -9.48e-07 * (central_latitude ** 4) + \
               -1.95e-09 * (central_latitude ** 5) + \
               6.15e-11 * (central_latitude ** 6)

    def norm_brdf(self,
                  data,
                  solar_za,
                  solar_az,
                  sensor_za,
                  sensor_az,
                  central_latitude=None,
                  sensor=None,
                  wavelengths=None,
                  nodata=0,
                  mask=None,
                  scale_factor=1.0,
                  out_range=None,
                  scale_angles=True):

        r"""
        Applies Nadir Bidirectional Reflectance Distribution Function (BRDF) normalization
        using the global c-factor method

        Args:
            data (2d or 3d DataArray): The data to normalize.
            solar_za (2d DataArray): The solar zenith angles (degrees).
            solar_az (2d DataArray): The solar azimuth angles (degrees).
            sensor_za (2d DataArray): The sensor azimuth angles (degrees).
            sensor_az (2d DataArray): The sensor azimuth angles (degrees).
            central_latitude (Optional[float or 2d DataArray]): The central latitude.
            sensor (Optional[str]): The satellite sensor.
            wavelengths (str list): The wavelength(s) to normalize.
            nodata (Optional[int or float]): A 'no data' value to fill NAs with.
            mask (Optional[DataArray]): A data mask, where clear values are 0.
            scale_factor (Optional[float]): A scale factor to apply to the input data.
            out_range (Optional[float]): The out data range. If not given, the output data are return in a 0-1 range.
            scale_angles (Optional[bool]): Whether to scale the pixel angle arrays.

        References:

            See :cite:`roy_etal_2016` for the c-factor method.

            For further background on BRDF:

                :cite:`li_strahler_1992`

                :cite:`roujean_etal_1992`

                :cite:`schaaf_etal_2002`

        Returns:
            ``xarray.DataArray``

        Examples:
            >>> import geowombat as gw
            >>> from geowombat.radiometry import BRDF
            >>>
            >>> brdf = BRDF()
            >>>
            >>> # Example where pixel angles are stored in separate GeoTiff files
            >>> with gw.config.update(sensor='l7', scale_factor=0.0001, nodata=0):
            >>>
            >>>     with gw.open('solarz.tif') as solarz,
            >>>         gw.open('solara.tif') as solara,
            >>>             gw.open('sensorz.tif') as sensorz,
            >>>                 gw.open('sensora.tif') as sensora:
            >>>
            >>>         with gw.open('landsat.tif') as src:
            >>>             src_norm = brdf.norm_brdf(src, solarz, solara, sensorz, sensora)
        """

        if not wavelengths:

            if sensor:
                wavelengths = list(data.gw.wavelengths[sensor]._fields)
            else:

                if not data.gw.sensor:
                    logger.exception('  The sensor must be supplied.')

                wavelengths = list(data.gw.wavelengths[data.gw.sensor]._fields)

        if not wavelengths:
            logger.exception('  The sensor or wavelength must be supplied.')

        if not isinstance(nodata, int) and not isinstance(nodata, float):
            nodata = data.gw.nodata

        # ne.set_num_threads(num_threads)

        if not isinstance(central_latitude, np.ndarray):
            if not isinstance(central_latitude, xr.DataArray):
                if not isinstance(central_latitude, float):

                    central_latitude = \
                    project_coords(np.array([data.x.values[int(data.x.shape[0] / 2)]], dtype='float64'),
                                   np.array([data.y.values[int(data.y.shape[0] / 2)]], dtype='float64'),
                                   data.crs,
                                   {'init': 'epsg:4326'})[1][0]

                    # TODO: rasterio.warp.reproject does not seem to be working
                    #
                    # Create the 2d latitudes
                    # central_latitude = project_coords(data.x.values,
                    #                                   data.y.values,
                    #                                   data.crs,
                    #                                   {'init': 'epsg:4326'},
                    #                                   num_threads=1,
                    #                                   warp_mem_limit=512)

        attrs = data.attrs.copy()

        if not nodata:
            nodata = data.gw.nodata

        if scale_factor == 1.0:
            scale_factor = data.gw.scale_factor

        # Scale the reflectance data
        if scale_factor != 1:
            data = data * scale_factor

        if scale_angles:

            # Scale the angle data to degrees

            solar_za = solar_za * 0.01
            solar_za.coords['band'] = [1]

            solar_az = solar_az * 0.01
            solar_az.coords['band'] = [1]

            sensor_za = sensor_za * 0.01
            sensor_za.coords['band'] = [1]

            sensor_az = sensor_az * 0.01
            sensor_az.coords['band'] = [1]

        # Get the Ross and Li coefficients
        self._get_kernels(central_latitude,
                          solar_za,
                          solar_az,
                          sensor_za,
                          sensor_az)

        # if len(wavelengths) == 1:
        #
        #     # Get the band iso, geo, and vol coefficients.
        #     coeffs = self.get_coeffs(wavelengths[0])
        #
        #     # Apply the adjustment.
        #     data = dask.delayed(ne.evaluate)(self.c_equation,
        #                                      local_dict=dict(fiso=coeffs['fiso'],
        #                                                      fgeo=coeffs['fgeo'],
        #                                                      fvol=coeffs['fvol'],
        #                                                      SA=data,
        #                                                      geo_norm=self.geo_norm,
        #                                                      geo_sensor=self.geo_sensor,
        #                                                      vol_norm=self.vol_norm,
        #                                                      vol_sensor=self.vol_sensor))
        #
        # else:

        results = list()

        for si, wavelength in enumerate(wavelengths):

            # Get the band iso, geo,
            #   and vol coefficients.
            coeffs = self._get_coeffs(wavelength)

            # c-factor
            c_factor = ((coeffs['fiso'] +
                         coeffs['fvol']*self.vol_norm +
                         coeffs['fgeo']*self.geo_norm) /
                        (coeffs['fiso'] +
                         coeffs['fvol']*self.vol_sensor +
                         coeffs['fgeo']*self.geo_sensor))

            p_norm = data.sel(band=wavelength).data * c_factor

            # Apply the adjustment to the current layer.
            results.append(p_norm)

        data = xr.DataArray(data=da.concatenate(results),
                            dims=('band', 'y', 'x'),
                            coords={'band': data.band.values,
                                    'y': data.y,
                                    'x': data.x},
                            attrs=data.attrs)

        if isinstance(out_range, float) or isinstance(out_range, int):

            if out_range <= 1:
                dtype = 'float64'
            elif 1 < out_range <= 255:
                dtype = 'uint8'
            else:
                dtype = 'uint16'

            drange = (0, out_range)

            data = (data * out_range).clip(0, out_range)

        else:

            drange = (0, 1)
            dtype = 'float64'

        # Mask data
        if isinstance(mask, xr.DataArray):
            data = xr.where((mask.sel(band=1) == 0) & (solar_za.sel(band=1) != -32768*0.01), data, nodata)
        else:
            data = xr.where(solar_za.sel(band=1) != -32768*0.01, data, nodata)

        data = data.transpose('band', 'y', 'x').astype(dtype)

        attrs['sensor'] = sensor
        attrs['calibration'] = 'BRDF-adjusted surface reflectance'
        attrs['nodata'] = nodata
        attrs['drange'] = drange

        data.attrs = attrs

        return data



# https://github.com/pydata/xarray/issues/2560
import netCDF4
import h5netcdf

import warnings

from contextlib import contextmanager

from . import geoxarray
from ..errors import logger
from ..backends import concat as gw_concat
from ..backends import mosaic as gw_mosaic
from ..backends import warp_open
from .util import Chunks, get_file_extension, parse_wildcard

import numpy as np
import xarray as xr
import rasterio as rio
from rasterio.windows import from_bounds, Window
import dask
import dask.array as da


warnings.filterwarnings('ignore')


ch = Chunks()

IO_DICT = dict(rasterio=['.tif',
                         '.tiff',
                         '.TIF',
                         '.TIFF',
                         '.img',
                         '.IMG',
                         '.vrt',
                         '.VRT',
                         '.jp2',
                         '.JP2',
                         '.hgt',
                         '.HGT'],
               xarray=['.nc'])


def get_attrs(src, **kwargs):

    cellxh = src.res[0] / 2.0
    cellyh = src.res[1] / 2.0

    left_ = src.bounds.left + (kwargs['window'].col_off * src.res[0]) + cellxh
    top_ = src.bounds.top - (kwargs['window'].row_off * src.res[1]) - cellyh

    xcoords = np.arange(left_, left_ + kwargs['window'].width * src.res[0], src.res[0])
    ycoords = np.arange(top_, top_ - kwargs['window'].height * src.res[1], -src.res[1])

    attrs = dict()

    attrs['transform'] = tuple(src.transform)[:6]

    if hasattr(src, 'crs') and src.crs:

        try:
            attrs['crs'] = src.crs.to_proj4()
        except:
            attrs['crs'] = src.crs.to_string()

    if hasattr(src, 'res'):
        attrs['res'] = src.res

    if hasattr(src, 'is_tiled'):
        attrs['is_tiled'] = np.uint8(src.is_tiled)

    if hasattr(src, 'nodatavals'):
        attrs['nodatavals'] = tuple(np.nan if nodataval is None else nodataval for nodataval in src.nodatavals)

    if hasattr(src, 'offsets'):
        attrs['offsets'] = src.scales

    if hasattr(src, 'offsets'):
        attrs['offsets'] = src.offsets

    if hasattr(src, 'descriptions') and any(src.descriptions):
        attrs['descriptions'] = src.descriptions

    if hasattr(src, 'units') and any(src.units):
        attrs['units'] = src.units

    return ycoords, xcoords, attrs


@dask.delayed
def read_delayed(fname, chunks, **kwargs):

    with rio.open(fname) as src:

        data_slice = src.read(**kwargs)

        single_band = True if len(data_slice.shape) == 2 else False

        if isinstance(chunks, int):
            chunks_ = (1, chunks, chunks)
        elif isinstance(chunks, tuple):
            chunks_ = (1,) + chunks if len(chunks) < 3 else chunks

        if single_band:

            # Expand to 1 band
            data_slice = da.from_array(data_slice[np.newaxis, :, :],
                                       chunks=chunks_)

        else:

            data_slice = da.from_array(data_slice,
                                       chunks=chunks)

        return data_slice


def read_list(file_list, chunks, **kwargs):
    return [read_delayed(fn, chunks, **kwargs) for fn in file_list]


def read(filename,
         band_names=None,
         time_names=None,
         bounds=None,
         chunks=256,
         num_workers=1,
         **kwargs):

    """
    Reads a window slice in-memory

    Args:
        filename (str or list): A file name or list of file names to open read.
        band_names (Optional[list]): A list of names to give the output band dimension.
        time_names (Optional[list]): A list of names to give the time dimension.
        bounds (Optional[1d array-like]): A bounding box to subset to, given as
            [minx, miny, maxx, maxy] or [left, bottom, right, top].
        chunks (Optional[tuple]): The data chunk size.
        num_workers (Optional[int]): The number of parallel ``dask`` workers.
        kwargs (Optional[dict]): Keyword arguments to pass to ``rasterio.write``.

    Returns:
        ``xarray.DataArray``
    """

    # Cannot pass 'chunks' to rasterio
    if 'chunks' in kwargs:
        del kwargs['chunks']

    if isinstance(filename, str):

        with rio.open(filename) as src:

            if bounds and ('window' not in kwargs):
                kwargs['window'] = from_bounds(*bounds, transform=src.transform)

            ycoords, xcoords, attrs = get_attrs(src, **kwargs)

        data = dask.compute(read_delayed(filename,
                                         chunks,
                                         **kwargs),
                            num_workers=num_workers)[0]

        if not band_names:
            band_names = np.arange(1, data.shape[0]+1)

        if len(band_names) != data.shape[0]:
            logger.exception('  The band names do not match the output dimensions.')
            raise ValueError

        data = xr.DataArray(data,
                            dims=('band', 'y', 'x'),
                            coords={'band': band_names,
                                    'y': ycoords,
                                    'x': xcoords},
                            attrs=attrs)

    else:

        with rio.open(filename[0]) as src:

            if bounds and ('window' not in kwargs):
                kwargs['window'] = from_bounds(*bounds, transform=src.transform)

            ycoords, xcoords, attrs = get_attrs(src, **kwargs)

        data = da.concatenate(dask.compute(read_list(filename,
                                                     chunks,
                                                     **kwargs),
                                           num_workers=num_workers),
                              axis=0)

        if not band_names:
            band_names = np.arange(1, data.shape[-3]+1)

        if len(band_names) != data.shape[-3]:
            logger.exception('  The band names do not match the output dimensions.')
            raise ValueError

        if not time_names:
            time_names = np.arange(1, len(filename)+1)

        if len(time_names) != data.shape[-4]:
            logger.exception('  The time names do not match the output dimensions.')
            raise ValueError

        data = xr.DataArray(data,
                            dims=('time', 'band', 'y', 'x'),
                            coords={'time': time_names,
                                    'band': band_names,
                                    'y': ycoords,
                                    'x': xcoords},
                            attrs=attrs)

    return data


@contextmanager
def open(filename,
         return_as='array',
         band_names=None,
         time_names=None,
         stack_dim='time',
         bounds=None,
         bounds_by='reference',
         resampling='nearest',
         mosaic=False,
         overlap='max',
         dtype=None,
         num_workers=1,
         **kwargs):

    """
    Opens a raster file

    Args:
        filename (str or list): The file name, search string, or a list of files to open.
        return_as (Optional[str]): The Xarray data type to return.
            Choices are ['array', 'dataset'] which correspond to ``xarray.DataArray`` and ``xarray.Dataset``.
        band_names (Optional[1d array-like]): A list of band names if ``return_as`` = 'dataset' or ``bounds``
            is given or ``window`` is given. Default is None.
        time_names (Optional[1d array-like]): A list of names to give the time dimension if ``bounds`` is given.
            Default is None.
        stack_dim (Optional[str]): The stack dimension. Choices are ['time', 'band'].
        bounds (Optional[1d array-like]): A bounding box to subset to, given as [minx, maxy, miny, maxx].
            Default is None.
        bounds_by (Optional[str]): How to concatenate the output extent if ``filename`` is a ``list`` and ``mosaic`` = ``False``.
            Choices are ['intersection', 'union', 'reference'].

            * reference: Use the bounds of the reference image. If a ``ref_image`` is not given, the first image in the ``filename`` list is used.
            * intersection: Use the intersection (i.e., minimum extent) of all the image bounds
            * union: Use the union (i.e., maximum extent) of all the image bounds

        resampling (Optional[str]): The resampling method if ``filename`` is a ``list``.
            Choices are ['average', 'bilinear', 'cubic', 'cubic_spline', 'gauss', 'lanczos', 'max', 'med', 'min', 'mode', 'nearest'].
        mosaic (Optional[bool]): If ``filename`` is a ``list``, whether to mosaic the arrays instead of stacking.
        overlap (Optional[str]): The keyword that determines how to handle overlapping data if ``filenames`` is a ``list``.
            Choices are ['min', 'max', 'mean'].
        dtype (Optional[str]): A data type to force the output to. If not given, the data type is extracted
            from the file.
        num_workers (Optional[int]): The number of parallel workers for Dask if ``bounds``
            is given or ``window`` is given. Default is 1.
        kwargs (Optional[dict]): Keyword arguments passed to the file opener.

    Returns:
        ``xarray.DataArray`` or ``xarray.Dataset``

    Examples:
        >>> import geowombat as gw
        >>>
        >>> # Open an image
        >>> with gw.open('image.tif') as ds:
        >>>     print(ds)
        >>>
        >>> # Open a list of images, stacking along the 'time' dimension
        >>> with gw.open(['image1.tif', 'image2.tif']) as ds:
        >>>     print(ds)
        >>>
        >>> # Open all GeoTiffs in a directory, stack along the 'time' dimension
        >>> with gw.open('*.tif') as ds:
        >>>     print(ds)
        >>>
        >>> # Use a context manager to handle images of difference sizes and projections
        >>> with gw.config.update(ref_image='image1.tif'):
        >>>
        >>>     # Use 'time' names to stack and mosaic non-aligned images with identical dates
        >>>     with gw.open(['image1.tif', 'image2.tif', 'image3.tif'],
        >>>
        >>>         # The first two images were acquired on the same date
        >>>         #   and will be merged into a single time layer
        >>>         time_names=['date1', 'date1', 'date2']) as ds:
        >>>
        >>>         print(ds)
        >>>
        >>> # Mosaic images across space using a reference
        >>> #   image for the CRS and cell resolution
        >>> with gw.config.update(ref_image='image1.tif'):
        >>>     with gw.open(['image1.tif', 'image2.tif'], mosaic=True) as ds:
        >>>         print(ds)
        >>>
        >>> # Mix configuration keywords
        >>> with gw.config.update(ref_crs='image1.tif', ref_res='image1.tif', ref_bounds='image2.tif'):
        >>>
        >>>     # The ``bounds_by`` keyword overrides the extent bounds
        >>>     with gw.open(['image1.tif', 'image2.tif'], bounds_by='union') as ds:
        >>>         print(ds)
        >>>
        >>> # Resample an image to 10m x 10m cell size
        >>> with gw.config.update(ref_crs=(10, 10)):
        >>>
        >>>     with gw.open('image.tif', resampling='cubic') as ds:
        >>>         print(ds)
        >>>
        >>> # Open a list of images at a window slice
        >>> from rasterio.windows import Window
        >>> w = Window(row_off=0, col_off=0, height=100, width=100)
        >>>
        >>> # Stack two images, opening band 3
        >>> with gw.open(['image1.tif', 'image2.tif'],
        >>>     band_names=['date1', 'date2'],
        >>>     num_workers=8,
        >>>     indexes=3,
        >>>     window=w,
        >>>     out_dtype='float32') as ds:
        >>>
        >>>     print(ds)
    """

    if return_as not in ['array', 'dataset']:
        logger.exception("  The `Xarray` object must be one of ['array', 'dataset']")

    if 'chunks' in kwargs:
        ch.check_chunktype(kwargs['chunks'], output='3d')

    if bounds or ('window' in kwargs and isinstance(kwargs['window'], Window)):

        if 'chunks' not in kwargs:

            if isinstance(filename, list):

                with rio.open(filename[0]) as src_:

                    w = src_.block_window(1, 0, 0)
                    chunks = (1, w.height, w.width)

            else:

                with rio.open(filename) as src_:

                    w = src_.block_window(1, 0, 0)
                    chunks = (1, w.height, w.width)

        else:
            chunks = kwargs['chunks']
            del kwargs['chunks']

        data = read(filename,
                    band_names=band_names,
                    time_names=time_names,
                    bounds=bounds,
                    chunks=chunks,
                    num_workers=num_workers,
                    **kwargs)

    else:

        if (isinstance(filename, str) and '*' in filename) or isinstance(filename, list):

            # Build the filename list
            if isinstance(filename, str):
                filename = parse_wildcard(filename)

            if 'chunks' not in kwargs:

                with rio.open(filename[0]) as src:

                    w = src.block_window(1, 0, 0)
                    kwargs['chunks'] = (1, w.height, w.width)

            if mosaic:

                # Mosaic images over space
                data = gw_mosaic(filename,
                                 overlap=overlap,
                                 bounds_by=bounds_by,
                                 resampling=resampling,
                                 band_names=band_names,
                                 dtype=dtype,
                                 **kwargs)

            else:

                # Stack images along the 'time' axis
                data = gw_concat(filename,
                                 stack_dim=stack_dim,
                                 bounds_by=bounds_by,
                                 resampling=resampling,
                                 time_names=time_names,
                                 band_names=band_names,
                                 overlap=overlap,
                                 dtype=dtype,
                                 **kwargs)

        else:

            file_names = get_file_extension(filename)

            if file_names.f_ext.lower() not in IO_DICT['rasterio'] + IO_DICT['xarray']:
                logger.exception('  The file format is not recognized.')

            if file_names.f_ext.lower() in IO_DICT['rasterio']:

                if 'chunks' not in kwargs:

                    with rio.open(filename) as src:

                        w = src.block_window(1, 0, 0)
                        kwargs['chunks'] = (1, w.height, w.width)

                data = warp_open(filename,
                                 band_names=band_names,
                                 resampling=resampling,
                                 dtype=dtype,
                                 **kwargs)

            else:

                if 'chunks' in kwargs and not isinstance(kwargs['chunks'], dict):
                    logger.exception('  The chunks should be a dictionary.')

                with xr.open_dataset(filename, **kwargs) as src:
                    data = src

    try:
        yield data
    finally:

        if hasattr(data.gw, '_obj'):
            data.gw._obj = None

        data.close()
        data = None



from .dask_ import Cluster
from .xarray_ import concat, mosaic
from .xarray_ import warp_open
from .xarray_ import to_crs
from .xarray_ import transform_crs

__all__ = ['Cluster',
           'concat',
           'mosaic',
           'warp_open',
           'to_crs',
           'transform_crs']



import os
from pathlib import Path
import fnmatch
import subprocess
from collections import namedtuple
import tarfile
# import datetime
# from datetime import datetime as dtime

from ..errors import logger

import numpy as np
import cv2
import xarray as xr
# import dask.array as da
import rasterio as rio
from rasterio.warp import reproject
from affine import Affine
import xml.etree.ElementTree as ET
# import numba as nb
# from pysolar.solar import get_altitude_fast, get_azimuth_fast


def scattering_angle(cos_sza, cos_vza, sin_sza, sin_vza, cos_raa):

    """
    Calculates the scattering angle

    Args:
        cos_sza (DataArray): The cosine of the solar zenith angle.
        cos_vza (DataArray): The cosine of the view zenith angle.
        sin_sza (DataArray): The sine of the solar zenith angle.
        sin_vza (DataArray): The sine of the view zenith angle.
        cos_raa (DataArray): The cosine of the relative azimuth angle.

    Equation:

        .. math::

            \Theta = scattering angle

            \theta_0 = solar zenith angle

            \theta_S = sensor zenith angle

            \zeta = relative azimuth angle

            \Theta_s = \arccos{- \cos{\theta_0} \cos{\theta_S} - \sin{\theta_0} \sin{\theta_S} \cos{\zeta}}

    References:
        scattering angle = the angle between the direction of incident and scattered radiation
        Liu, CH and Liu GR (2009) AEROSOL OPTICAL DEPTH RETRIEVAL FOR SPOT HRV IMAGES, Journal of Marine Science and Technology
        http://stcorp.github.io/harp/doc/html/algorithms/derivations/scattering_angle.html

    Returns:
        Scattering angle (in radians) as an ``xarray.DataArray``
    """

    scattering_angle = xr.ufuncs.arccos(-cos_sza * cos_vza - sin_sza * sin_vza * cos_raa)

    return xr.ufuncs.cos(scattering_angle) ** 2


def relative_azimuth(saa, vaa):

    """
    Calculates the relative azimuth angle

    Args:
        saa (DataArray): The solar azimuth angle (in degrees).
        vaa (DataArray): The view azimuth angle (in degrees).

    Reference:
        http://stcorp.github.io/harp/doc/html/algorithms/derivations/relative_azimuth_angle.html

    Returns:
        Relative azimuth (in degrees) as an ``xarray.DataArray``
    """

    # Relative azimuth (in radians)
    raa = xr.ufuncs.deg2rad(saa - vaa)

    # Create masks
    raa_plus = xr.where(raa >= 2.0*np.pi, 1, 0)
    raa_minus = xr.where(raa < 0, 1, 0)

    raa = xr.where(raa_plus == 1, raa + (2.0*np.pi), raa)
    raa = xr.where(raa_minus == 1, raa - (2.0*np.pi), raa)

    return xr.ufuncs.rad2deg(raa)


def get_sentinel_sensor(metadata):

    # Parse the XML file
    tree = ET.parse(metadata)
    root = tree.getroot()

    for child in root:

        if 'general_info' in child.tag[-14:].lower():
            general_info = child

    for ginfo in general_info:

        if ginfo.tag == 'TILE_ID':
            file_name = ginfo.text

    return file_name[:3].lower()


def _parse_sentinel_angles(metadata, proc_angles, nodata):

    """
    Gets the Sentinel-2 solar angles from metadata

    Args:
        metadata (str): The metadata file.
        proc_angles (str): The angles to parse. Choices are ['solar', 'view'].
        nodata (int or float): The 'no data' value.

    Returns:
        zenith and azimuth angles as a ``tuple`` of 2d ``numpy`` arrays
    """

    if proc_angles == 'view':

        zenith_values = np.zeros((13, 23, 23), dtype='float64') + nodata
        azimuth_values = np.zeros((13, 23, 23), dtype='float64') + nodata

    else:

        zenith_values = np.zeros((23, 23), dtype='float64') + nodata
        azimuth_values = np.zeros((23, 23), dtype='float64') + nodata

    view_tag = 'Sun_Angles_Grid' if proc_angles == 'solar' else 'Viewing_Incidence_Angles_Grids'

    # Parse the XML file
    tree = ET.parse(metadata)
    root = tree.getroot()

    # Find the angles
    for child in root:

        if child.tag[-14:] == 'Geometric_Info':
            geoinfo = child

    for segment in geoinfo:

        if segment.tag == 'Tile_Angles':
            angles = segment

    for angle in angles:

        if angle.tag == view_tag:

            if proc_angles == 'view':
                band_id = int(angle.attrib['bandId'])

            for bset in angle:

                if bset.tag == 'Zenith':
                    zenith = bset
                if bset.tag == 'Azimuth':
                    azimuth = bset

            for field in zenith:

                if field.tag == 'Values_List':
                    zvallist = field

            for field in azimuth:

                if field.tag == 'Values_List':
                    avallist = field

            for rindex in range(len(zvallist)):

                zvalrow = zvallist[rindex]
                avalrow = avallist[rindex]
                zvalues = zvalrow.text.split(' ')
                avalues = avalrow.text.split(' ')
                values = list(zip(zvalues, avalues))

                for cindex in range(len(values)):

                    if (values[cindex][0].lower() != 'nan') and (values[cindex][1].lower() != 'nan'):

                        ze = float(values[cindex][0])
                        az = float(values[cindex][1])

                        if proc_angles == 'view':

                            zenith_values[band_id, rindex, cindex] = ze
                            azimuth_values[band_id, rindex, cindex] = az

                        else:

                            zenith_values[rindex, cindex] = ze
                            azimuth_values[rindex, cindex] = az

    return zenith_values, azimuth_values


def sentinel_pixel_angles(metadata,
                          ref_file,
                          outdir='.',
                          nodata=-32768,
                          overwrite=False,
                          verbose=0):

    """
    Generates Sentinel pixel angle files

    Args:
        metadata (str): The metadata file.
        ref_file (str): A reference image to use for geo-information.
        outdir (Optional[str])): The output directory to save the angle files to.
        nodata (Optional[int or float]): The 'no data' value.
        overwrite (Optional[bool]): Whether to overwrite existing angle files.
        verbose (Optional[int]): The verbosity level.

    References:
        https://www.sentinel-hub.com/faq/how-can-i-access-meta-data-information-sentinel-2-l2a
        https://github.com/marujore/sentinel_angle_bands/blob/master/sentinel2_angle_bands.py

    Returns:
        zenith and azimuth angles as a ``namedtuple`` of angle file names
    """

    AngleInfo = namedtuple('AngleInfo', 'vza vaa sza saa sensor')

    sza, saa = _parse_sentinel_angles(metadata, 'solar', nodata)
    vza, vaa = _parse_sentinel_angles(metadata, 'view', nodata)

    sensor_name = get_sentinel_sensor(metadata)

    with rio.open(ref_file) as src:

        profile = src.profile.copy()

        ref_height = src.height
        ref_width = src.width
        ref_extent = src.bounds

        profile.update(transform=Affine(src.res[0], 0.0, ref_extent.left, 0.0, -src.res[1], ref_extent.top),
                       height=ref_height,
                       width=ref_width,
                       nodata=-32768,
                       dtype='int16',
                       count=1,
                       driver='GTiff',
                       tiled=True,
                       compress='lzw')

    ref_base = '_'.join(os.path.basename(ref_file).split('_')[:-1])

    opath = Path(outdir)

    opath.mkdir(parents=True, exist_ok=True)

    # Set output angle file names.
    sensor_azimuth_file = opath.joinpath(ref_base + '_sensor_azimuth.tif').as_posix()
    sensor_zenith_file = opath.joinpath(ref_base + '_sensor_zenith.tif').as_posix()
    solar_azimuth_file = opath.joinpath(ref_base + '_solar_azimuth.tif').as_posix()
    solar_zenith_file = opath.joinpath(ref_base + '_solar_zenith.tif').as_posix()

    for angle_array, angle_file in zip([vaa,
                                        vza,
                                        saa,
                                        sza],
                                       [sensor_azimuth_file,
                                        sensor_zenith_file,
                                        solar_azimuth_file,
                                        solar_zenith_file]):

        pfile = Path(angle_file)

        if overwrite:

            if pfile.is_file():
                pfile.unlink()

        if not pfile.is_file():

            # TODO: write data for each band?
            if len(angle_array.shape) > 2:
                angle_array = angle_array.mean(axis=0)

            with rio.open(angle_file, mode='w', **profile) as dst:

                if verbose > 0:
                    logger.info('  Writing {} to file ...'.format(angle_file))

                # Resample and scale
                angle_array_resamp = np.int16(cv2.resize(angle_array,
                                                         (0, 0),
                                                         fy=ref_height / angle_array.shape[0],
                                                         fx=ref_width / angle_array.shape[1],
                                                         interpolation=cv2.INTER_LINEAR) / 0.01)

                dst.write(angle_array_resamp, indexes=1)

    return AngleInfo(vaa=sensor_azimuth_file,
                     vza=sensor_zenith_file,
                     saa=solar_azimuth_file,
                     sza=solar_zenith_file,
                     sensor=sensor_name)


# Potentially useful for angle creation
# https://github.com/gee-community/gee_tools/blob/master/geetools/algorithms.py

# def slope_between(a, b):
#     return (a[1] - b[1]) / (a[0] - b[0])
#
#
# @nb.jit
# def _calc_sensor_angles(data,
#                         zenith_angles,
#                         azimuth_angles,
#                         yvalues,
#                         xvalues,
#                         celly,
#                         cellx,
#                         satellite_height,
#                         nodata,
#                         acquisition_date):
#
#     """
#     Calculates sensor zenith and azimuth angles
#     """
#
#     slope = slope_between(np.array([data.gw.meta.right + ((data.gw.ncols/2.0)*data.gw.cellx), data.gw.meta.top]),
#                           np.array([data.gw.meta.left, data.gw.meta.top - ((data.gw.nrows / 2.0) * data.gw.celly)]))
#
#     slope_perc = -1.0 / slope
#
#     view_az = (math.pi / 2.0) - math.arctan(slope_perc)
#
#     for i in range(0, yvalues.shape[0]):
#
#         for j in range(0, xvalues.shape[0]):
#
#             if data_band[i, j] != nodata:
#
#                 # TODO: calculate satellite drift angle
#                 dist_from_nadir = None
#
#                 # Calculate the distance from the current location to the satellite
#                 dist_to_satellite = np.hypot(satellite_height, dist_from_nadir)
#
#                 # Calculate the view angle
#
#                 zenith_angles[i, j]
#
#                 # Solar zenith angle = 90 - elevation angle scaled to integer range
#                 zenith_angles[i, j] = (90.0 - get_altitude_fast(xvalues[j], yvalues[i], acquisition_date)) / 0.01
#
#                 # Solar azimuth angle
#                 azimuth_angles[i, j] = float(get_azimuth_fast(xvalues[j], yvalues[i], acquisition_date)) / 0.01
#
#     return zenith_angles, azimuth_angles


# @nb.jit
# def _calc_solar_angles(data_band, zenith_angles, azimuth_angles, yvalues, xvalues, nodata, acquisition_date):
#
#     """
#     Calculates solar zenith and azimuth angles
#     """
#
#     for i in range(0, yvalues):
#
#         for j in range(0, xvalues):
#
#             if data_band[i, j] != nodata:
#
#                 # Solar zenith angle = 90 - elevation angle scaled to integer range
#                 zenith_angles[i, j] = (90.0 - get_altitude_fast(xvalues[j], yvalues[i], acquisition_date)) / 0.01
#
#                 # Solar azimuth angle
#                 azimuth_angles[i, j] = float(get_azimuth_fast(xvalues[j], yvalues[i], acquisition_date)) / 0.01
#
#     return zenith_angles, azimuth_angles


# def pixel_angles(data, band, nodata, meta):
#
#     """
#     Generates pixel zenith and azimuth angles
#
#     Args:
#         data (Xarray): The data with coordinate and transform attributes.
#         band (int or str): The ``data`` band to use for masking.
#         nodata (int or float): The 'no data' value in ``data``.
#         meta (namedtuple): The metadata file. Should have image acquisition year, month, day and hour attributes.
#     """
#
#     acquisition_date = dtime(meta.year, meta.month, meta.day, meta.hour, 0, 0, 0, tzinfo=datetime.timezone.utc)
#
#     yvalues = data.y.values
#     xvalues = data.x.values
#
#     data_band = data.sel(band=band).data.compute()
#     sze = np.zeros((data.gw.nrows, data.gw.ncols), dtype='int16') - 32768
#     saa = np.zeros((data.gw.nrows, data.gw.ncols), dtype='int16') - 32768
#
#     sze, saa = _calc_solar_angles(data_band, sze, saa, yvalues, xvalues, nodata, acquisition_date)
#
#     sze_attrs = data.attrs.copy()
#     saa_attrs = data.attrs.copy()
#
#     sze_attrs['values'] = 'Solar zenith angle'
#     sze_attrs['scale_factor'] = 0.01
#
#     saa_attrs['values'] = 'Solar azimuth angle'
#     sze_attrs['scale_factor'] = 0.01
#
#     szex = xr.DataArray(data=da.from_array(sze[np.newaxis, :, :],
#                                            chunks=(1, data.gw.row_chunks, data.gw.col_chunks)),
#                         coords={'band': 'sze',
#                                 'y': data.y,
#                                 'x': data.x},
#                         dims=('band', 'y', 'x'),
#                         attrs=sze_attrs)
#
#     saax = xr.DataArray(data=da.from_array(saa[np.newaxis, :, :],
#                                            chunks=(1, data.gw.row_chunks, data.gw.col_chunks)),
#                         coords={'band': 'saa',
#                                 'y': data.y,
#                                 'x': data.x},
#                         dims=('band', 'y', 'x'),
#                         attrs=saa_attrs)
#
#     return szex, saax


def landsat_pixel_angles(angles_file,
                         ref_file,
                         outdir,
                         sensor,
                         l57_angles_path=None,
                         l8_angles_path=None,
                         verbose=0):

    """
    Generates Landsat pixel angle files

    Args:
        angles_file (str): The angles file.
        ref_file (str): A reference file.
        outdir (str): The output directory.
        sensor (str): The sensor.
        l57_angles_path (str): The path to the Landsat 5 and 7 angles bin.
        l8_angles_path (str): The path to the Landsat 8 angles bin.
        verbose (Optional[int]): The verbosity level.

    Returns:
        zenith and azimuth angles as a ``namedtuple`` of angle file names
    """

    if not l57_angles_path:

        gw_bin = os.path.realpath(os.path.dirname(__file__))

        gw_out = os.path.realpath(Path(gw_bin).joinpath('../bin').as_posix())
        gw_tar = os.path.realpath(Path(gw_bin).joinpath('../bin/ESPA.tar.gz').as_posix())

        if not Path(gw_bin).joinpath('../bin/ESPA').is_dir():

            with tarfile.open(gw_tar, mode='r:gz') as tf:
                tf.extractall(gw_out)

        l57_angles_path = Path(gw_out).joinpath('ESPA/landsat_angles').as_posix()
        l8_angles_path = Path(gw_out).joinpath('ESPA/l8_angles').as_posix()

    AngleInfo = namedtuple('AngleInfo', 'vza vaa sza saa')

    # Setup the angles name.
    # example file = LE07_L1TP_225098_20160911_20161008_01_T1_sr_band1.tif

    with rio.open(ref_file) as src:

        ref_height = src.height
        ref_width = src.width
        ref_extent = src.bounds

    ref_base = '_'.join(os.path.basename(ref_file).split('_')[:-1])

    opath = Path(outdir)

    opath.mkdir(parents=True, exist_ok=True)

    # Set output angle file names.
    sensor_azimuth_file = opath.joinpath(ref_base + '_sensor_azimuth.tif').as_posix()
    sensor_zenith_file = opath.joinpath(ref_base + '_sensor_zenith.tif').as_posix()
    solar_azimuth_file = opath.joinpath(ref_base + '_solar_azimuth.tif').as_posix()
    solar_zenith_file = opath.joinpath(ref_base + '_solar_zenith.tif').as_posix()

    if not Path(sensor_azimuth_file).is_file():

        # Setup the command.
        if sensor.lower() in ['l5', 'l7']:

            angle_command = '{PATH} {META} -s 1 -b 1'.format(PATH=Path(l57_angles_path).joinpath('landsat_angles').as_posix(),
                                                             META=angles_file)

            # 1=zenith, 2=azimuth
            out_order = dict(azimuth=2, zenith=1)
            # out_order = [2, 1, 2, 1]

        else:

            angle_command = '{PATH} {META} BOTH 1 -f -32768 -b 4'.format(PATH=Path(l8_angles_path).joinpath('l8_angles').as_posix(),
                                                                         META=angles_file)

            # 1=azimuth, 2=zenith
            out_order = dict(azimuth=1, zenith=2)
            # out_order = [1, 2, 1, 2]

        os.chdir(outdir)

        if verbose > 0:
            logger.info('  Generating pixel angles ...')

        # Create the angle files.
        subprocess.call(angle_command, shell=True)

        # Get angle data from 1 band.
        sensor_angles = fnmatch.filter(os.listdir(outdir), '*sensor_B04.img')[0]
        solar_angles = fnmatch.filter(os.listdir(outdir), '*solar_B04.img')[0]

        sensor_angles_fn_in = opath.joinpath(sensor_angles).as_posix()
        solar_angles_fn_in = opath.joinpath(solar_angles).as_posix()

        # Convert the data
        for in_angle, out_angle, band_pos in zip([sensor_angles_fn_in,
                                                  sensor_angles_fn_in,
                                                  solar_angles_fn_in,
                                                  solar_angles_fn_in],
                                                 [sensor_azimuth_file,
                                                  sensor_zenith_file,
                                                  solar_azimuth_file,
                                                  solar_zenith_file],
                                                 ['azimuth',
                                                  'zenith',
                                                  'azimuth',
                                                  'zenith']):

            with rio.open(in_angle) as src:

                profile = src.profile.copy()

                profile.update(transform=Affine(src.res[0], 0.0, ref_extent.left, 0.0, -src.res[1], ref_extent.top),
                               height=ref_height,
                               width=ref_width,
                               nodata=-32768,
                               dtype='int16',
                               count=1,
                               driver='GTiff',
                               tiled=True,
                               compress='lzw')

                src_band = rio.Band(src, out_order[band_pos], 'int16', (src.height, src.width))

                with rio.open(out_angle, mode='w', **profile) as dst:

                    dst_band = rio.Band(dst, 1, 'int16', (dst.height, dst.width))

                    # TODO: num_threads
                    reproject(src_band,
                              destination=dst_band,
                              num_threads=8)

    return AngleInfo(vaa=sensor_azimuth_file,
                     vza=sensor_zenith_file,
                     saa=solar_azimuth_file,
                     sza=solar_zenith_file)



import math
from collections import namedtuple
from datetime import datetime as dtime
import datetime

from .angles import relative_azimuth

import numpy as np
import pandas as pd
import xarray as xr
import dask.array as da


def p_r(m, r, rphase, cos_solar_za, cos_sensor_za):

    """
    Calculates atmospheric reflectance due to Rayleigh scattering

    Args:
        m (float): The air mass.
        r (float): The Rayleigh optical depth.
        rphase (float): The Rayleigh phase function.
        cos_solar_za (DataArray): The cosine of the solar zenith angle.
        cos_sensor_za (DataArray): The cosine of the sensor zenith angle.

    Returns:
        ``xarray.DataArray``
    """

    rphase_stack = xr.concat([rphase] * len(m.band), dim='band')
    rphase_stack.coords['band'] = m.band.values

    cos_solar_za_stack = xr.concat([cos_solar_za] * len(m.band), dim='band')
    cos_solar_za_stack.coords['band'] = m.band.values

    cos_sensor_za_stack = xr.concat([cos_sensor_za] * len(m.band), dim='band')
    cos_sensor_za_stack.coords['band'] = m.band.values

    return rphase_stack * ((1.0 - xr.ufuncs.exp(-m*r)) / (4.0 * (cos_solar_za_stack + cos_sensor_za_stack)))


def t_sv(r, cos_zenith):

    """
    Calculates atmospheric transmittance of sun-surface path

    Args:
        r (float): The Rayleigh optical depth.
        cos_zenith (DataArray): The cosine of the zenith angle.

    Returns:
        ``xarray.DataArray``
    """

    cos_zenith_stack = xr.concat([cos_zenith] * len(r.band), dim='band')
    cos_zenith_stack.coords['band'] = r.band.values

    cose1 = xr.ufuncs.exp(-r / cos_zenith_stack)
    cose2 = xr.ufuncs.exp(0.52*r / cos_zenith_stack)

    return cose1 + cose1 * (cose2 - 1.0)


def s_atm(r):

    """
    Calculates atmospheric backscattering ratio to count multiple reflections between the surface and atmosphere

    Args:
        r (float): The Rayleigh optical depth.

    Returns:
        ``float``
    """

    return (0.92*r) * xr.ufuncs.exp(-r)


def _format_coeff(dataframe, sensor, key):

    bands_dict = dict(l5={'1': 'blue', '2': 'green', '3': 'red', '4': 'nir', '5': 'swir1', '6': 'swir2'},
                      l7={'1': 'blue', '2': 'green', '3': 'red', '4': 'nir', '5': 'swir1', '6VCID1': 'th1',
                          '6VCID2': 'th2', '7': 'swir2', '8': 'pan'},
                      l8={'1': 'coastal', '2': 'blue', '3': 'green', '4': 'red', '5': 'nir', '6': 'swir1',
                          '7': 'swir2', '8': 'pan', '9': 'cirrus', '10': 'th1', '11': 'th2'})

    sensor_dict = bands_dict[sensor]

    dataframe_ = dataframe[dataframe.iloc[:, 0].str.startswith(key)].values

    pairs = dict()

    for di in range(dataframe_.shape[0]):

        bd = dataframe_[di, 0]
        cf = dataframe_[di, 1]

        try:
            pairs[sensor_dict[''.join(bd.split('_')[3:])]] = float(cf)
        except:
            pass

    # dataframe_[:, 1] = dataframe_[:, 1].astype(float)
    # dataframe_[:, 0] = list(range(1, dataframe_.shape[0]+1))

    return pairs


class MetaData(object):

    """
    A class for sensor metadata
    """

    @staticmethod
    def get_landsat_coefficients(meta_file):

        """
        Gets coefficients from a Landsat metadata file

        Args:
            meta_file (str): A metadata file.

        Returns:

            ``namedtuple``:

                sensor, m_l, a_l, m_p, a_p
        """

        associations = {'LANDSAT_5': 'l5',
                        'LANDSAT_7': 'l7',
                        'LANDSAT_8': 'l8'}

        MetaCoeffs = namedtuple('MetaCoeffs', 'sensor m_l a_l m_p a_p date_acquired')

        df = pd.read_csv(meta_file, sep='=')

        df.iloc[:, 0] = df.iloc[:, 0].str.strip()
        df.iloc[:, 1] = df.iloc[:, 1].str.strip()

        spacecraft_id = dict(df[df.iloc[:, 0].str.startswith('SPACECRAFT_ID')].values)
        spacecraft_id['SPACECRAFT_ID'] = spacecraft_id['SPACECRAFT_ID'].replace('"', '')
        sensor = associations[spacecraft_id['SPACECRAFT_ID']]

        m_l = _format_coeff(df, sensor, 'RADIANCE_MULT_BAND_')
        a_l = _format_coeff(df, sensor, 'RADIANCE_ADD_BAND_')
        m_p = _format_coeff(df, sensor, 'REFLECTANCE_MULT_BAND_')
        a_p = _format_coeff(df, sensor, 'REFLECTANCE_ADD_BAND_')

        date_acquired_ = dict(df[df.iloc[:, 0].str.startswith('DATE_ACQUIRED')].values)
        date_acquired_ = date_acquired_['DATE_ACQUIRED'].replace('"', '')
        year, month, day = date_acquired_.split('-')

        year = int(year)
        month = int(month)
        day = int(day)

        scene_center_time = dict(df[df.iloc[:, 0].str.startswith('SCENE_CENTER_TIME')].values)
        scene_center_time = scene_center_time['SCENE_CENTER_TIME'].replace('"', '')
        hour = int(scene_center_time.split(':')[0])

        date_acquired = dtime(year, month, day, hour, tzinfo=datetime.timezone.utc)

        return MetaCoeffs(sensor=sensor,
                          m_l=m_l,
                          a_l=a_l,
                          m_p=m_p,
                          a_p=a_p,
                          date_acquired=date_acquired)


class LinearAdjustments(object):

    """
    A class for linear bandpass adjustments
    """

    def __init__(self):

        self.coefficients = dict(s2a=dict(l5=None,
                                          l7=None,
                                          l8=dict(alphas=dict(coastal=-0.0002,
                                                              blue=-0.004,
                                                              green=-0.0009,
                                                              red=0.0009,
                                                              nir=-0.0001,
                                                              swir1=-0.0011,
                                                              swir2=-0.0012),
                                                  betas=dict(coastal=0.9959,
                                                             blue=0.9778,
                                                             green=1.0053,
                                                             red=0.9765,
                                                             nir=0.9983,
                                                             swir1=0.9987,
                                                             swir2=1.003))),
                                 s2b=dict(l5=None,
                                          l7=None,
                                          l8=dict(alphas=dict(coastal=-0.0002,
                                                              blue=-0.004,
                                                              green=-0.0008,
                                                              red=0.001,
                                                              nir=0.0,
                                                              swir1=-0.0003,
                                                              swir2=0.0004),
                                                  betas=dict(coastal=0.9959,
                                                             blue=0.9778,
                                                             green=1.0075,
                                                             red=0.9761,
                                                             nir=0.9966,
                                                             swir1=1.0,
                                                             swir2=0.9867))),
                                 l7=dict(l5=None,
                                         l8=dict(alphas=dict(blue=-0.0095,
                                                             green=-0.0016,
                                                             red=-0.0022,
                                                             nir=-0.0021,
                                                             swir1=-0.003,
                                                             swir2=0.0029,
                                                             pan=-0.00443),
                                                 betas=dict(blue=0.9785,
                                                            green=0.9542,
                                                            red=0.9825,
                                                            nir=1.0073,
                                                            swir1=1.0171,
                                                            swir2=0.9949,
                                                            pan=0.9717)),
                                         s2=None))

    def bandpass(self,
                 data,
                 sensor=None,
                 to='l8',
                 band_names=None,
                 scale_factor=1):

        """
        Applies a bandpass adjustment by applying a linear function to surface reflectance values

        Args:
            data (DataArray): The data to adjust.
            sensor (Optional[str]): The sensor to adjust.
            to (Optional[str]): The sensor to adjust to.
            band_names (Optional[list]): The bands to adjust. If not given, all bands are adjusted.
            scale_factor (Optional[float]): A scale factor to apply to the input data.

        Reference:

            Sentinel-2 and Landsat 8:

                https://hls.gsfc.nasa.gov/algorithms/bandpass-adjustment/

                See :cite:`chastain_etal_2019` for further details

            Landsat 7 and Landsat 8:

                See :cite:`roy_etal_2016` (Table 2)

        Returns:
            ``xarray.DataArray``

        Examples:
            >>> import geowombat as gw
            >>> from geowombat.radiometry import LinearAdjustments
            >>>
            >>> la = LinearAdjustments()
            >>>
            >>> # Adjust all Sentinel-2 bands to Landsat 8
            >>> with gw.config.update(sensor='s2'):
            >>>     with gw.open('sentinel-2.tif') as ds:
            >>>         ds_adjusted = la.bandpass(ds, to='l8')
        """

        attrs = data.attrs.copy()

        if scale_factor == 1.0:
            scale_factor = data.gw.scale_factor

        # Scale the reflectance data
        if scale_factor != 1:
            data = data * scale_factor

        if not band_names:

            band_names = data.band.values.tolist()

            if band_names[0] == 1:
                band_names = list(data.gw.wavelengths[sensor]._fields)

        coeff_dict = self.coefficients[sensor][to]

        alphas = np.array([coeff_dict['alphas'][bd] for bd in band_names], dtype='float64')
        betas = np.array([coeff_dict['betas'][bd] for bd in band_names], dtype='float64')

        alphas = xr.DataArray(data=alphas,
                              coords={'band': band_names},
                              dims='band')

        betas = xr.DataArray(data=betas,
                             coords={'band': band_names},
                             dims='band')

        # Apply the linear bandpass adjustment
        data = alphas + betas * data

        if scale_factor != 1:
            data = data / scale_factor

        data.attrs['adjustment'] = '{} to {}'.format(sensor, to)
        data.attrs['alphas'] = alphas.data.tolist()
        data.attrs['betas'] = betas.data.tolist()

        data.attrs = attrs

        return data


class RadTransforms(MetaData):

    """
    A class for radiometric transformations
    """

    def dn_to_sr(self,
                 dn,
                 solar_za,
                 solar_az,
                 sensor_za,
                 sensor_az,
                 nodata=-32768,
                 sensor=None,
                 method='srem',
                 angle_factor=0.01,
                 meta=None):

        """
        Converts digital numbers to surface reflectance

        Args:
            dn (DataArray): The digital number data to calibrate.
            solar_za (DataArray): The solar zenith angle.
            solar_az (DataArray): The solar azimuth angle.
            sensor_za (DataArray): The sensor, or view, zenith angle.
            sensor_az (DataArray): The sensor, or view, azimuth angle.
            nodata (Optional[int or float]): The 'no data' value from the pixel angle data.
            sensor (Optional[str]): The data's sensor.
            method (Optional[str]): The method to use. Only 'srem' is supported.
            angle_factor (Optional[float]): The scale factor for angles.
            meta (Optional[namedtuple]): A metadata object with gain and bias coefficients.

        References:
            https://www.usgs.gov/land-resources/nli/landsat/using-usgs-landsat-level-1-data-product

        Returns:

            ``xarray.DataArray``:

                Data range: 0-1

        Examples:
            >>> from geowombat.radiometry import RadTransforms
            >>>
            >>> sr = RadTransforms()
            >>> meta = sr.get_landsat_coefficients('file.MTL')
            >>>
            >>> # Convert DNs to surface reflectance using Landsat metadata
            >>> with gw.open('dn.tif') as ds:
            >>>     sr_data = sr.dn_to_sr(ds, solar_za, sensor_za, meta=meta)
        """

        attrs = dn.attrs.copy()

        # Get the data band names and positional indices
        band_names = dn.band.values.tolist()

        if meta:

            if not sensor:
                sensor = meta.sensor

            # Get the gain and offsets and
            #   convert the gain and offsets
            #   to named coordinates.
            m_p = xr.DataArray(data=[meta.m_p[bi] for bi in band_names], coords={'band': band_names}, dims='band')
            a_p = xr.DataArray(data=[meta.a_p[bi] for bi in band_names], coords={'band': band_names}, dims='band')

            toar = self.dn_to_toar(dn, m_p, a_p)

            # TOAR with sun angle correction
            cos_sza = xr.concat([xr.ufuncs.cos(xr.ufuncs.deg2rad(solar_za*angle_factor))] * len(toar.band), dim='band')
            cos_sza.coords['band'] = toar.band.values
            toar = toar / cos_sza
            toar.attrs = attrs

        else:

            if not sensor:
                sensor = dn.gw.sensor

            # d = distance between the Earth and Sun in the astronomical unit
            # ESUN = mean solar exoatmospheric radiation
            GlobalArgs = namedtuple('GlobalArgs', 'pi d esun')

            # TODO: set global arguments
            global_args = GlobalArgs(pi=math.pi, d=None, esun=None)

            radiance = self.dn_to_radiance(dn, None, None)

            toar = self.radiance_to_toar(radiance, solar_za, global_args)

        sr_data = self.toar_to_sr(toar,
                                  solar_za,
                                  solar_az,
                                  sensor_za,
                                  sensor_az,
                                  sensor,
                                  nodata=nodata)

        sr_data = sr_data.where(sr_data != nodata)

        attrs['sensor'] = sensor
        attrs['nodata'] = nodata
        attrs['calibration'] = 'surface reflectance'
        attrs['method'] = method
        attrs['drange'] = (0, 1)

        sr_data.attrs = attrs

        return sr_data

    @staticmethod
    def dn_to_radiance(dn, gain, bias):

        """
        Converts digital numbers to radiance

        Args:
            dn (DataArray): The digital number data to calibrate.
            gain (Optional[float]): A gain value.
            bias (Optional[float]): A bias value.

        Returns:
            ``xarray.DataArray``
        """

        attrs = dn.attrs.copy()

        # TODO: get gain and bias from metadata
        rad_data = gain * dn + bias

        attrs['calibration'] = 'radiance'

        rad_data.attrs = attrs

        return rad_data

    @staticmethod
    def dn_to_toar(dn, gain, bias):

        """
        Converts digital numbers to top-of-atmosphere reflectance

        Args:
            dn (DataArray): The digital number data to calibrate.
            gain (Optional[float]): A gain value.
            bias (Optional[float]): A bias value.

        Returns:
            ``xarray.DataArray``
        """

        attrs = dn.attrs.copy()

        # TODO: get gain and bias from metadata
        toar_data = gain * dn + bias

        attrs['calibration'] = 'top-of-atmosphere reflectance'

        toar_data.attrs = attrs

        return toar_data

    @staticmethod
    def radiance_to_toar(radiance, solar_za, global_args):

        """
        Converts digital numbers to top-of-atmosphere reflectance

        Args:
            radiance (DataArray): The radiance data to calibrate.
            solar_za (DataArray): The solar zenith angle.
            global_args (namedtuple): Global arguments.

        Returns:
            ``xarray.DataArray``
        """

        attrs = radiance.attrs.copy()

        solar_zenith_angle = solar_za * 0.01

        toar_data = (global_args.pi * radiance * global_args.d**2) / (global_args.esun * da.cos(solar_zenith_angle))

        attrs['calibration'] = 'top-of-atmosphere reflectance'

        toar_data.attrs = attrs

        return toar_data

    @staticmethod
    def toar_to_sr(toar,
                   solar_za,
                   solar_az,
                   sensor_za,
                   sensor_az,
                   sensor,
                   nodata=-32768):

        """
        Converts top-of-atmosphere reflectance to surface reflectance

        Args:
            toar (DataArray): The top-of-atmosphere reflectance.
            solar_za (DataArray): The solar zenith angle.
            solar_az (DataArray): The solar azimuth angle.
            sensor_za (DataArray): The sensor zenith angle.
            sensor_az (DataArray): The sensor azimuth angle.
            sensor (str): The satellite sensor.
            nodata (Optional[int or float]): The 'no data' value from the pixel angle data.

        References:

            See :cite:`bilal_etal_2019`

        Returns:

            ``xarray.DataArray``:

                Data range: 0-1
        """

        attrs = toar.attrs.copy()

        central_um = toar.gw.central_um[sensor]
        band_names = list(toar.gw.wavelengths[sensor]._fields)
        band_um = [getattr(central_um, p) for p in band_names]
        um = xr.DataArray(data=band_um, coords={'band': band_names}, dims='band')

        # Scale the angles to degrees
        sza = solar_za * 0.01
        sza.coords['band'] = [1]

        saa = solar_az * 0.01
        saa.coords['band'] = [1]

        vza = sensor_za * 0.01
        vza.coords['band'] = [1]

        vaa = sensor_az * 0.01
        vaa.coords['band'] = [1]

        # Convert to radians
        rad_sza = xr.ufuncs.deg2rad(sza)
        rad_vza = xr.ufuncs.deg2rad(vza)

        # Cosine(deg2rad(angles)) = angles x (pi / 180)
        cos_sza = xr.ufuncs.cos(rad_sza)
        cos_sza.coords['band'] = [1]
        cos_vza = xr.ufuncs.cos(rad_vza)
        cos_vza.coords['band'] = [1]

        sin_sza = xr.ufuncs.sin(rad_sza)
        sin_sza.coords['band'] = [1]
        sin_vza = xr.ufuncs.sin(rad_vza)
        sin_vza.coords['band'] = [1]

        # air mass
        m = (1.0 / cos_sza.sel(band=1)) + (1.0 / cos_vza.sel(band=1))
        m = m.expand_dims(dim='band')
        m = m.assign_coords(band=[1])

        m = xr.concat([m]*len(toar.band), dim='band')
        m.coords['band'] = toar.band.values

        # Rayleigh optical depth
        # Hansen, JF and Travis, LD (1974) LIGHT SCATTERING IN PLANETARY ATMOSPHERES
        r = 0.008569*um**-4 * (1.0 + 0.0113*um**-2 + 0.0013*um**-4)

        # Relative azimuth angle
        # TODO: doesn't work if the band coordinate is named
        raa = relative_azimuth(saa, vaa)
        rad_raa = xr.ufuncs.deg2rad(raa)
        cos_raa = xr.ufuncs.cos(rad_raa)

        # scattering angle = the angle between the direction of incident and scattered radiation
        # Liu, CH and Liu GR (2009) AEROSOL OPTICAL DEPTH RETRIEVAL FOR SPOT HRV IMAGES, Journal of Marine Science and Technology
        # http://stcorp.github.io/harp/doc/html/algorithms/derivations/scattering_angle.html
        scattering_angle = xr.ufuncs.arccos(-cos_sza * cos_vza - sin_sza * sin_vza * cos_raa)
        cos2_scattering_angle = xr.ufuncs.cos(scattering_angle)**2

        # Rayleigh phase function
        rphase = ((3.0 * 0.9587256) / (4.0 + 1.0 - 0.9587256)) * (1.0 + cos2_scattering_angle)

        pr_data = p_r(m, r, rphase, cos_sza, cos_vza)

        # da.nan_to_num(pr_data).max().compute()

        toar_diff = toar - pr_data

        # Total transmission = downward x upward
        transmission = t_sv(r, cos_sza) * t_sv(r, cos_vza)

        # Atmospheric backscattering ratio
        ab_ratio = s_atm(r)

        sr_data = (toar_diff / (toar_diff * ab_ratio + transmission)).fillna(nodata).clip(0, 1).astype('float64')

        attrs['sensor'] = sensor
        attrs['calibration'] = 'surface reflectance'
        attrs['nodata'] = nodata
        attrs['drange'] = (0, 1)

        sr_data.attrs = attrs

        return sr_data
